from ._NDice import NDice

class Dice6(NDice):
    """
    Simple dice with 6 sides

    __init__():
        int seed --> seed for the random generator (deafult is None)
    """
    def __init__(self, seed = None):
        super.__init__(6, seed)


from collections.abc import MutableSequence
from ._NDice import NDice

class DiceCup(MutableSequence):
    """
    Collection of NDice objects that are rolled together

    Atributes:
        list AllDice --> list of the NDice objects rolled by the cup
        int Count --> number of dice stored
        bool Dynamic --> whether or not the cup is mutable

    __init__():
        list sides --> list of side numbers representing unique dice. For multiple of the same number, use a single element and alter noOfDice
        int noOfDice --> number of dice to create (deafult is length of sides list)
        bool dynamic --> whether or not the cup is mutable
    """
    def __init__(self, sides: list, noOfDice: int = None, dynamic: bool = False):
        #? super().__init__()

        self.Dynamic = dynamic
        self.AllDice: list = NDice.MultipleDice(sides, noOfDice)

        self.Count: int = property(fget = lambda self: len(self.AllDice))

        if not dynamic and len(self.AllDice) == 0:
            raise ValueError("No dice were specified for a static dice cup.")

    def Roll(self, suppressEmptyExeption = False):
        """
        Rolls all the dice and returns each individual result
        """
        if not suppressEmptyExeption and self.Count == 0:
            raise ValueError("There are no dice in the cup.");

        results = []

        for dice in self.AllDice:
            results.append(dice.Roll())

        return results

    def RollTotal(self, suppressEmptyExeption = False):
        """
        Rolls all the dice and returns the sum of the results
        """
        return sum(self.Roll(suppressEmptyExeption))

    def __getitem__(self, key):
        if not self.Dynamic:
            raise RuntimeError("Invalid operation - the dice cup is not dynamic!")

        return self.AllDice.__getitem__(key)

    def __setitem__(self, key, value):
        if not self.Dynamic:
            raise RuntimeError("Invalid operation - the dice cup is not dynamic!")

        self.AllDice.__setitem__(key, value)

    def __delitem__(self, key):
        if not self.Dynamic:
            raise RuntimeError("Invalid operation - the dice cup is not dynamic!")

        self.AllDice.__delitem__(key)

    def __len__(self):
        return len(self.AllDice)

    def insert(self, index, object):
        if not self.Dynamic:
            raise RuntimeError("Invalid operation - the dice cup is not dynamic!")

        self.AllDice.insert(index, object)

    def __iadd__(self, value):
        self.append(value)
        return self


from typing import Dict

from ._Char import Char

# \odot
odot: Char = Char("\u2299")

# Superscripts
#---------------------------------------------------------------

# Superscript 0
Superscript_0: Char = Char("\u2070")

# Superscript 1
Superscript_1: Char = Char("\u00B9")

# Superscript 2
Superscript_2: Char = Char("\u00B2")

# Superscript 3
Superscript_3: Char = Char("\u00B3")

# Superscript 4
Superscript_4: Char = Char("\u2074")

# Superscript 5
Superscript_5: Char = Char("\u2075")

# Superscript 6
Superscript_6: Char = Char("\u2076")

# Superscript 7
Superscript_7: Char = Char("\u2077")

# Superscript 8
Superscript_8: Char = Char("\u2078")

# Superscript 9
Superscript_9: Char = Char("\u2079")

# Superscript i
Superscript_i: Char = Char("\u2071")

# Superscript +
Superscript_plus: Char = Char("\u207A")

# Superscript -
Superscript_minus: Char = Char("\u207B")

# Superscript =
Superscript_equals: Char = Char("\u207C")

# Superscript (
Superscript_bracket_left: Char = Char("\u207D")

# Superscript )
Superscript_bracket_right: Char = Char("\u207E")

# Superscript lookups
superscripts: Dict[Char, Char] = {
    Char("0"): Superscript_0,
    Char("1"): Superscript_1,
    Char("2"): Superscript_2,
    Char("3"): Superscript_3,
    Char("4"): Superscript_4,
    Char("5"): Superscript_5,
    Char("6"): Superscript_6,
    Char("7"): Superscript_7,
    Char("8"): Superscript_8,
    Char("9"): Superscript_9,
    Char("i"): Superscript_i,
    Char("+"): Superscript_plus,
    Char("-"): Superscript_minus,
    Char("="): Superscript_equals,
    Char("("): Superscript_bracket_left,
    Char(")"): Superscript_bracket_right,
}
superscripts_reverse: Dict[Char, Char] = {
    Superscript_0: Char("0"),
    Superscript_1: Char("1"),
    Superscript_2: Char("2"),
    Superscript_3: Char("3"),
    Superscript_4: Char("4"),
    Superscript_5: Char("5"),
    Superscript_6: Char("6"),
    Superscript_7: Char("7"),
    Superscript_8: Char("8"),
    Superscript_9: Char("9"),
    Superscript_i: Char("i"),
    Superscript_plus: Char("+"),
    Superscript_minus: Char("-"),
    Superscript_equals: Char("="),
    Superscript_bracket_left: Char("("),
    Superscript_bracket_right: Char(")"),
}

def to_superscript(value: str) -> str:
    return Char.to_string([superscripts[c] for c in Char.from_string(value)])

def from_superscript(value: str) -> str:
    return Char.to_string([superscripts_reverse[c] for c in Char.from_string(value)])

# Subscripts
#---------------------------------------------------------------

# Subscript 0
Subscript_0: Char = Char("\u2080")

# Subscript 1
Subscript_1: Char = Char("\u2081")

# Subscript 2
Subscript_2: Char = Char("\u2082")

# Subscript 3
Subscript_3: Char = Char("\u2083")

# Subscript 4
Subscript_4: Char = Char("\u2084")

# Subscript 5
Subscript_5: Char = Char("\u2085")

# Subscript 6
Subscript_6: Char = Char("\u2086")

# Subscript 7
Subscript_7: Char = Char("\u2087")

# Subscript 8
Subscript_8: Char = Char("\u2088")

# Subscript 9
Subscript_9: Char = Char("\u2089")

# Subscript +
Subscript_plus: Char = Char("\u208A")

# Subscript -
Subscript_minus: Char = Char("\u208B")

# Subscript =
Subscript_equals: Char = Char("\u208C")

# Subscript (
Subscript_bracket_left: Char = Char("\u208D")

# Subscript )
Subscript_bracket_right: Char = Char("\u208E")

# Subscript lookups
subscripts: Dict[Char, Char] = {
    Char("0"): Subscript_0,
    Char("1"): Subscript_1,
    Char("2"): Subscript_2,
    Char("3"): Subscript_3,
    Char("4"): Subscript_4,
    Char("5"): Subscript_5,
    Char("6"): Subscript_6,
    Char("7"): Subscript_7,
    Char("8"): Subscript_8,
    Char("9"): Subscript_9,
    Char("+"): Subscript_plus,
    Char("-"): Subscript_minus,
    Char("="): Subscript_equals,
    Char("("): Subscript_bracket_left,
    Char(")"): Subscript_bracket_right,
}
subscripts_reverse: Dict[Char, Char] = {
    Subscript_0: Char("0"),
    Subscript_1: Char("1"),
    Subscript_2: Char("2"),
    Subscript_3: Char("3"),
    Subscript_4: Char("4"),
    Subscript_5: Char("5"),
    Subscript_6: Char("6"),
    Subscript_7: Char("7"),
    Subscript_8: Char("8"),
    Subscript_9: Char("9"),
    Subscript_plus: Char("+"),
    Subscript_minus: Char("-"),
    Subscript_equals: Char("="),
    Subscript_bracket_left: Char("("),
    Subscript_bracket_right: Char(")"),
}

def to_subscript(value: str) -> str:
    return Char.to_string([subscripts[c] for c in Char.from_string(value)])

def from_subscript(value: str) -> str:
    return Char.to_string([subscripts_reverse[c] for c in Char.from_string(value)])



import yaml

from ._ConfigsBase import ConfigsBase

class YamlConfig(ConfigsBase):
    def __init__(self, writable: bool = True, filepath: str = None, safe_load = True, expand_maps: bool = True, map_expansion_limit: int = -1):
        if map_expansion_limit < -1 or map_expansion_limit == 0:
            raise ValueError(f"paramiter map_expansion_limit cannot have a value of {map_expansion_limit}. It must be either a positive, non-zero integer, or -1 to indicate no limit.")
        self.__safe_load = safe_load
        self.__expand_maps = expand_maps
        self.__map_expansion_limit = map_expansion_limit
        super().__init__(writable = writable, filepath = filepath)

    def read(self, filepath: str):
        with open(filepath, "r") as file:
            if self.__safe_load:
                yaml_data = yaml.safe_load(file)
            else:
                yaml_data = yaml.unsafe_load(file)

            def recursivley_add_mappings(root, data, n_further_expansions):
                for key in data:
                    if n_further_expansions != 0 and isinstance(data[key], dict):
                        root.add_group(key)
                        recursivley_add_mappings(root[key], data[key], n_further_expansions if n_further_expansions == -1 else n_further_expansions - 1)
                    else:
                        root.add_item(key, data[key])

            recursivley_add_mappings(self, yaml_data, self.__map_expansion_limit if self.__expand_maps else 0)

    @staticmethod
    def from_file(filepath: str, writable: bool = False, safe_load = True, expand_maps: bool = True, map_expansion_limit: int = -1):
        return YamlConfig(writable, filepath, safe_load, expand_maps, map_expansion_limit)



from collections.abc import Callable
from functools import wraps
from typing import TypeVar, TypeAlias, Literal

from matplotlib import pyplot as plt
from matplotlib.collections import PolyCollection
from matplotlib.axes import Axes
from matplotlib.colors import Colormap
from matplotlib.typing import ColorType
import numpy as np

from ..Data._Rect import Rect
from ._CachedPlotElements import CachedPlotHexbin

COLOUR_FUNCTION_TYPE: TypeAlias = Callable[[float|None, float|None, float|None], Callable[[np.ndarray[tuple[int], np.dtype[np.integer]]], float]]

class Hexbin(object):

    COLOUR_FUNCTION_TYPE = COLOUR_FUNCTION_TYPE
    
    def __init__(
        self,
        x_data: np.ndarray[tuple[int], np.dtype[np.floating]],
        y_data: np.ndarray[tuple[int], np.dtype[np.floating]],
        colour_function: COLOUR_FUNCTION_TYPE = None,
        cache: CachedPlotHexbin|None = None
    ) -> None:
        super().__init__()
        self.__x_data = x_data
        self.__y_data = y_data
        self.__colour_function = colour_function if colour_function is not None else Hexbin.create_hexbin_count()
        self.__cache_object = cache if cache is not None else CachedPlotHexbin()

    @property
    def colour_function(self) -> COLOUR_FUNCTION_TYPE: # type: ignore[valid-type]
        """
        The function used to generate another function for calculating the colour of the hexbin cells.
        """
        return self.__colour_function

    @property
    def data(self) -> CachedPlotHexbin:
        """
        Get the cache object for this hexbin.
        """
        return self.__cache_object
    
    def plot_hexbin(
        self,
        default_bin_value: float = -np.inf,
        extent: Rect|None = None,
        gridsize: int|tuple[int, int]|None = None,
        colourmap: str|Colormap|None = None,
        min_value: float|None = None,
        max_value: float|None = None,
        edge_colour: ColorType|Literal["face", "none"]|None = None,
        alpha_values: np.ndarray[tuple[int], np.dtype[np.floating]]|None = None,
        axis: Axes|None = None,
        **hexbin_kwargs
    ) -> PolyCollection:
        """
        Plot a hexbin using a pre-defined colour scheme.

        hexbin_kwargs will be passed to the `plt.hexbin` function call (or that of the provided axis).

        May raise ValueError if any of the arguments not provided are not set on the data object.
        """

        extent = extent if extent is not None else self.data.extent
        gridsize = gridsize if gridsize is not None else self.data.gridsize
        colourmap = colourmap if colourmap is not None else self.data.colourmap
        min_value = min_value if min_value is not None else self.data.min_value
        max_value = max_value if max_value is not None else self.data.max_value
        edge_colour = edge_colour if edge_colour is not None else self.data.edgecolour
        alpha_values = alpha_values if alpha_values is not None else self.data.bin_alphas

        hexes = (axis if axis is not None else plt).hexbin(
            x = self.__x_data,
            y = self.__y_data,
            C = np.arange(len(self.__x_data)),
            reduce_C_function = self.__colour_function(min_value, max_value, default_bin_value),
            extent = extent.extent,
            gridsize = gridsize,
            cmap = colourmap,
            vmin = min_value,
            vmax = max_value,
            edgecolor = edge_colour,
            **hexbin_kwargs
        )

        if alpha_values is not None:
            hexes.set_alpha(alpha_values)

        self.__cache_object = CachedPlotHexbin.from_hexes(hexes, extent, gridsize, min_value, max_value, edgecolours_match_faces = edge_colour == "face")

        return hexes



    @staticmethod
    def create_hexbin_colour_function(statistic: Callable[[np.ndarray[tuple[int], np.dtype[np.integer]]], float]):
        """
        Create a function for calculating the value for colour mapping of hexbin cells.

        The return value of this function can be passed to `plot_hexbin` as an argument for the parameter `colour_function`.
        """

        def ready_hexbin_colour_function(min_value: float|None = None, max_value: float|None = None, default_bin_value: float = -np.inf):

            @wraps(statistic)
            def calculate_bin_colour(indices: np.ndarray, /) -> float:
                if len(indices) == 0:
                    return default_bin_value
                try:
                    result = statistic(indices)
                except Exception:
                    return default_bin_value
                if min_value is not None and result < min_value:
                    return min_value
                elif max_value is not None and result > max_value:
                    return max_value
                else:
                    return result

            return calculate_bin_colour
        
        return ready_hexbin_colour_function



    @staticmethod
    def create_hexbin_count():
        """
        Assign hexbin cell colour to the number of elements that fall within the bin.
        """
        @Hexbin.create_hexbin_colour_function
        def calculate_statistic_count(indices: np.ndarray, /) -> float:
            return len(indices)
        return calculate_statistic_count



    @staticmethod
    def create_hexbin_log10_count():
        """
        Assign hexbin cell colour to the log_10 of the number of elements that fall within the bin.
        """
        @Hexbin.create_hexbin_colour_function
        def calculate_statistic_log10_count(indices: np.ndarray, /) -> float:
            return np.log10(len(indices))
        return calculate_statistic_log10_count



    @staticmethod
    def create_hexbin_fraction(mask: np.ndarray[tuple[int], np.dtype[np.bool_]]):
        """
        Assign hexbin cell colour to the fraction of of elements that fall within the bin.
        """
        @Hexbin.create_hexbin_colour_function
        def calculate_statistic_fraction(indices: np.ndarray, /) -> float:
            return mask[indices].sum() / len(indices)
        return calculate_statistic_fraction



    @staticmethod
    def create_hexbin_quantity_fraction(data: np.ndarray[tuple[int], np.dtype[np.floating]], mask: np.ndarray[tuple[int], np.dtype[np.bool_]]):
        """
        Assign hexbin cell colour to the fraction of of elements that fall within the bin.
        """
        @Hexbin.create_hexbin_colour_function
        def calculate_statistic_quantity_fraction(indices: np.ndarray, /) -> float:
            subset = data[indices]
            return subset[mask[indices]].sum() / subset.sum()
        return calculate_statistic_quantity_fraction



    @staticmethod
    def create_hexbin_sum(data: np.ndarray):
        """
        Assign hexbin cell colour to the sum of the elements from this dataset that fall within the bin.
        """
        @Hexbin.create_hexbin_colour_function
        def calculate_statistic_sum(indices: np.ndarray, /) -> float:
            return data[indices].sum()
        return calculate_statistic_sum



    @staticmethod
    def create_hexbin_log10_sum(data: np.ndarray):
        """
        Assign hexbin cell colour to the log_10 of the sum of the elements from this dataset that fall within the bin.
        """
        @Hexbin.create_hexbin_colour_function
        def calculate_statistic_log10_sum(indices: np.ndarray, /) -> float:
            return np.log10(data[indices].sum())
        return calculate_statistic_log10_sum



    @staticmethod
    def create_hexbin_sum_log10(data: np.ndarray):
        """
        Assign hexbin cell colour to the sum of log_10 of the the elements from this dataset that fall within the bin.

        Note, it is advised you use `create_hexbin_sum` and pass the `np.log10(data)` to it instead of using this function, unless retaining a second copy of the data array is problematic.
        """
        @Hexbin.create_hexbin_colour_function
        def calculate_statistic_sum_log10(indices: np.ndarray, /) -> float:
            return np.log10(data[indices]).sum()
        return calculate_statistic_sum_log10



    @staticmethod
    def create_hexbin_mean(data: np.ndarray, offset: float = 0.0):
        """
        Assign hexbin cell colour to the mean of elements from this dataset that fall within the bin.
        """
        @Hexbin.create_hexbin_colour_function
        def calculate_statistic_mean(indices: np.ndarray, /) -> float:
            return np.mean(data[indices]) + offset
        return calculate_statistic_mean



    @staticmethod
    def create_hexbin_log10_mean(data: np.ndarray, offset: float = 0.0):
        """
        Assign hexbin cell colour to the mean of elements from this dataset that fall within the bin.
        """
        @Hexbin.create_hexbin_colour_function
        def calculate_statistic_log10_mean(indices: np.ndarray, /) -> float:
            return np.log10(np.mean(data[indices])) + offset
        return calculate_statistic_log10_mean



    @staticmethod
    def create_hexbin_weighted_mean(data: np.ndarray, weights: np.ndarray, offset: float = 0.0):
        """
        Assign hexbin cell colour to the mean of elements from this dataset that fall within the bin.

        Use the weights from an array of equal length.
        """
        @Hexbin.create_hexbin_colour_function
        def calculate_statistic_weighted_mean(indices: np.ndarray, /) -> float:
            return np.average(data[indices], weights = weights[indices]) + offset
        return calculate_statistic_weighted_mean



    @staticmethod
    def create_hexbin_log10_weighted_mean(data: np.ndarray, weights: np.ndarray, offset: float = 0.0):
        """
        Assign hexbin cell colour to the mean of elements from this dataset that fall within the bin.

        Use the weights from an array of equal length.
        """
        @Hexbin.create_hexbin_colour_function
        def calculate_statistic_log10_weighted_mean(indices: np.ndarray, /) -> float:
            return np.log10(np.average(data[indices], weights = weights[indices])) + offset
        return calculate_statistic_log10_weighted_mean



    @staticmethod
    def create_hexbin_median(data: np.ndarray):
        """
        Assign hexbin cell colour to the median of elements from this dataset that fall within the bin.
        """
        @Hexbin.create_hexbin_colour_function
        def calculate_statistic_median(indices: np.ndarray, /) -> float:
            return np.median(data[indices])
        return calculate_statistic_median



    @staticmethod
    def create_hexbin_percentile(data: np.ndarray, percentile: float):
        """
        Assign hexbin cell colour to the specified percentile of elements from this dataset that fall within the bin.
        """
        @Hexbin.create_hexbin_colour_function
        def calculate_statistic_percentile(indices: np.ndarray, /) -> float:
            return np.percentile(data[indices], percentile)
        return calculate_statistic_percentile



from typing import Union, List, Tuple, Iterable, Any

class Char(str):
    def __new__(cls, object: Union["Char", str]):
        if not isinstance(object, str):
            raise TypeError(f"{type(object)} is neither Char or str.")
        if len(object) != 1:
            raise ValueError(f"{object} is not a valid character or character string.")
        str.__new__(cls, object)

    @staticmethod
    def from_string(value: str) -> Tuple["Char", ...]:
        """
        Create a sequence of characters from a string.
        """
        return tuple(Char(c) for c in value)

    @staticmethod
    def to_string(*chars: Union["Char", str, Iterable[Union["Char", str]]]) -> str:
        """
        Create a string from a sequence of characters.
        """
        return "".join([c if isinstance(c, str) and len(c) == 1 else Char.to_string(*c) for c in chars])



from ._VersionInformation import VersionInfomation
from ._PersonInfomation import PersonInfomation
from ._AuthorInfomation import AuthorInfomation
from ._Rect import IRect, GenericRect, Rect, DiscreteRect
from ._Rect3D import IRect3D, GenericRect3D, Rect3D, DiscreteRect3D



from scipy import stats
from scipy.optimize import least_squares
import math
import matplotlib.pyplot as plt
import numpy as np
from enum import Enum

IAutomaticFigure = object# Partial initialisation to allow circular reference for type hints. See EOF for import
from .data import error_string, standard_form_string_if_nessessary

class LineType(Enum):
    """
    An enumeration listing the valid line types used by the LineOfBestFit class.
    """

    none = 0
    straight = 1,# mx + c
    quadratic = 2,# ax^2 + bx + c
    cubic = 3,# ax^3 + bx^2 + cx + d
    quartic = 4,# ax^4 + bx^3 + cx^2 + dx + e
    power = 5,# ax^b + c
    exponential = 6,# a e^(bx) + c
    logarithmic = 7,# a e^(bx) + c where b is negitive
    sin = 8,# a sin(bx + c)
    #gaussian = 9,# TODO: add this
    custom = 9


class LineOfBestFit(object):
    """
    Calculates and stores a line of best fit using an associated Figure object holding the data and optional errors.

    Arguments:
        Figure figure -> the accociated Figure object with the data needed to calculate the fit.

        LineType lineType -> The type of line to be created. Use LineType.custom for custom options.
                                (Defualt is LineType.straight)

        dict initialFitVariables -> A dictionary containing the fit variables and their starting value used to
                                    overide he defults or to specify a custom fit. Variables should be in oposite
                                    order to that expected by the fitting functions i.e. m then c for y = m * x + c (Defult is None)

        list customFitMethods -> A list with two items: [the fitting function, the error function] (Defult is None)

    A custom fit can be produced by specifying the lineType argument as LineType.custom and then specifying the fit
    variables and providing the fitting functions. LineOfBestFit.createGenericErrorFunc allows a fit function to be
    specified to produce a generic error function. Resulting customFitMethods for this would be
    [fitFunc, errorFunc] or [fitFunc, LineOfBestFit.createGenericErrorFunc(fitFunc)]
    """

    def __init__(self, figure: IAutomaticFigure, lineType: LineType = LineType.straight, initialFitVariables: dict = None, customFitMethods: list = None):
        self.lineType = lineType
        self.figure: Figure = figure

        if initialFitVariables is None:
            if self.lineType == LineType.straight:
                self.initialFitVariables = {"m": 1.0, "c": 1.0}
            elif self.lineType == LineType.quadratic:
                self.initialFitVariables = {"a": 1.0, "b": 1.0, "c": 1.0}
            elif self.lineType == LineType.cubic:
                self.initialFitVariables = {"a": 1.0, "b": 1.0, "c": 1.0, "d": 1.0}
            elif self.lineType == LineType.quartic:
                self.initialFitVariables = {"a": 1.0, "b": 1.0, "c": 1.0, "d": 1.0, "e": 1.0}
            elif self.lineType == LineType.power:
                self.initialFitVariables = {"a": 1.0, "b": 1.0, "c": 1.0}
            elif self.lineType == LineType.exponential:
                self.initialFitVariables = {"a": 1.0, "b": 1.0, "c": 1.0}
            elif self.lineType == LineType.logarithmic:
                self.initialFitVariables = {"a": 1.0, "b": -1.0, "c": 1.0}
            elif self.lineType == LineType.sin:
                self.initialFitVariables = {"a": 1.0, "b": 1.0, "c": 1.0}
            elif self.lineType == LineType.custom:
                raise ValueError("No fit variables were provided. This is required for a custom fit.")
        else:
            self.initialFitVariables = initialFitVariables

        if customFitMethods is None:
            if self.lineType == LineType.straight:
                self.fitLine = LineOfBestFit.fitLine_straight
                self.fitError = LineOfBestFit.fitError_straight
            elif self.lineType == LineType.quadratic:
                self.fitLine = LineOfBestFit.fitLine_quadratic
                self.fitError = LineOfBestFit.fitError_quadratic
            elif self.lineType == LineType.cubic:
                self.fitLine = LineOfBestFit.fitLine_cubic
                self.fitError = LineOfBestFit.fitError_cubic
            elif self.lineType == LineType.quartic:
                self.fitLine = LineOfBestFit.fitLine_quartic
                self.fitError = LineOfBestFit.fitError_quartic
            elif self.lineType == LineType.power:
                self.fitLine = LineOfBestFit.fitLine_power
                self.fitError = LineOfBestFit.fitError_power
            elif self.lineType == LineType.exponential or lineType == LineType.logarithmic:
                self.fitLine = LineOfBestFit.fitLine_exponential
                self.fitError = LineOfBestFit.fitError_exponential
            elif self.lineType == LineType.sin:
                self.fitLine = LineOfBestFit.fitLine_sin
                self.fitError = LineOfBestFit.fitError_sin
            elif self.lineType == LineType.custom:
                raise ValueError("No fit functions were provided. This is required for a custom fit.")
        else:
            if not isinstance(customFitMethods, list):
                raise TypeError("The value provided for the argument of customFitMethods was not a list.")
            elif len(customFitMethods) != 2:
                raise ValueError("The value provided for the argument of customFitMethods had {} elements. It must have exactly 2.".format(len(customFitMethods)))

            self.fitLine = customFitMethods[0]
            self.fitError = customFitMethods[1]

        self.fit_X = None
        self.fit_Y = None
        self.fittingOutput = None
        self.fitVariables = {}
        self.fitQuality = None

        self.runFit()

    def runFit(self, linePoints = 1000000):
        """
        Runs the fitting algorithem with the current fit atributes

        Arguments:
            int linePoints -> the number of x values to use within the min to max range of
                                the data points in order to create the fit line. (Defult is 1,000,000)

        Credits:
            Tim Greenshaw
            Christopher Rowe
        """

        if not isinstance(linePoints, int):
            raise TypeError("The value provided for the argument \"linePoints\" was not of type int.")
        elif linePoints <= 0:
            raise ValueError("The number of points must be a non zero positive value.")

        if not isinstance(self.figure, IAutomaticFigure):
            raise AttributeError("The \"figure\" property was not set as no valid figure has been linked.")

        initialValues = []
        self.fitVariables = {}

        # In the event that the line type none was specified - don't create a line.
        if self.lineType == LineType.none or self.lineType is None:
            self.fittingOutput = "Fit type \"none\" specified. No fit generated."
            self.fitQuality = math.inf
            for key in self.initialFitVariables:
                self.fitVariables[key] = None
            self.fit_X = None
            self.fit_Y = None
            return

        # Set the initial values for the specified type of line
        elif self.lineType == LineType.straight:
            initialValues = [self.initialFitVariables["c"], self.initialFitVariables["m"]]
        elif self.lineType == LineType.quadratic:
            initialValues = [self.initialFitVariables["c"], self.initialFitVariables["b"], self.initialFitVariables["a"]]
        elif self.lineType == LineType.cubic:
            initialValues = [self.initialFitVariables["d"], self.initialFitVariables["c"], self.initialFitVariables["b"], self.initialFitVariables["a"]]
        elif self.lineType == LineType.quartic:
            initialValues = [self.initialFitVariables["e"], self.initialFitVariables["d"], self.initialFitVariables["c"], self.initialFitVariables["b"], self.initialFitVariables["a"]]
        elif self.lineType == LineType.power:
            initialValues = [self.initialFitVariables["c"], self.initialFitVariables["b"], self.initialFitVariables["a"]]
        elif self.lineType == LineType.exponential or self.lineType == LineType.logarithmic:
            initialValues = [self.initialFitVariables["c"], self.initialFitVariables["b"], self.initialFitVariables["a"]]
        elif self.lineType == LineType.sin:
            initialValues = [self.initialFitVariables["c"], self.initialFitVariables["b"], self.initialFitVariables["a"]]
        elif self.lineType == LineType.custom:
            initialValues = [self.initialFitVariables[key] for key in self.initialFitVariables]
            initialValues.reverse()


        numberOfPoints = np.size(self.figure.get_x())
        out = least_squares(self.fitError, initialValues, args=(self.figure.get_x(), self.figure.get_y(), self.figure.get_xError() if self.figure.get_xError() is not None else 0, self.figure.get_yError() if self.figure.get_yError() is not None else 0))

        if not out.success:
            #raise RuntimeWarning("Fit failed.")
            self.fittingOutput = "Fit failed."
            self.fitQuality = math.inf
            self.fitVariables = {}
            for key in self.initialFitVariables:
                self.fitVariables[key] = None
            self.fit_X = None
            self.fit_Y = None
            return
            

        output = ""

        finalValues = out.x

        # Calculate chi^2 per point, summed chi^2 and chi^2/NDF (Number of Degrees of Freedom)
        chiList = self.fitError(finalValues, self.figure.get_x(), self.figure.get_y(), self.figure.get_xError() if self.figure.get_xError() is not None else 0, self.figure.get_yError() if self.figure.get_yError() is not None else 0)**2
        chiSquared = np.sum(chiList)
        NDF = numberOfPoints - 2#TODO: check this - overide option?
        reducedChiSquared = chiSquared/NDF

        output += "\nFit quality:" \
                + "\n    chi Squared per point:\n        " + str(chiList) \
                + "\n\n    chi Squared = {}".format(standard_form_string_if_nessessary(chiSquared, 5)) \
                + "\n    chi Squared / NDF = {}".format(standard_form_string_if_nessessary(reducedChiSquared, 5)) \
                + "\n\n    Final Fit Variables:\n        " + str(finalValues) \
                + "\n\n    y:\n        " + str(self.figure.get_y()) \
                + "\n\n    dx:\n        " + str(self.figure.get_xError()) \
                + "\n\n    dy:\n        " + str(self.figure.get_yError())

        # Compute covariance
        jMat = out.jac
        jMat2 = np.dot(jMat.T, jMat)
        detJmat2 = np.linalg.det(jMat2)

        finalErrors = []
        if detJmat2 < 1E-32:
            for i in range(len(finalValues)):
                finalErrors.append(0.0)

            output += "\n\nValue of determinat detJmat2 " + str(detJmat2) + "\nMatrix singular, error calculation failed."

        else:
            covar = np.linalg.inv(jMat2)
            for i in range(len(finalValues)):
                finalErrors.append(np.sqrt(covar[i, i]))
            
        # Calculate fitted function values
        self.fit_X = np.linspace(min(self.figure.get_x()), max(self.figure.get_x()), linePoints)
        self.fit_Y = self.fitLine(finalValues, self.fit_X)
        
        self.fitQuality = reducedChiSquared

        lastIndex = len(finalValues) - 1
        for i, key in enumerate(self.initialFitVariables):
            self.fitVariables[key] = (finalValues[lastIndex - i], finalErrors[lastIndex - i])

        output += "\n\nParameters returned by fit:"

        for key in self.fitVariables:
            output += "\n    " + str(key) + " = " + error_string(self.fitVariables[key][0], self.fitVariables[key][1])

        self.fittingOutput = output

    @staticmethod
    def fitLine_straight(p: list, x):
        """
        Calculates a y value for a straight line
        y_guess = mx + c

        p: [c, m]

        Credits:
            Tim Greenshaw
            Christopher Rowe
        """
        f = p[1]*x + p[0]
        return f
    
    @staticmethod
    def fitError_straight(p: list, x, y, xerr, yerr):
        """
        Calculates the error for a straight line
        error = (y - y_guess) / sqrt(dy^2 + m^2 * dx^2)

        p: [c, m]

        Credits:
            Tim Greenshaw
            Christopher Rowe
        """
        #if not (isinstance(xerr, int) and xerr == 0) and not (isinstance(yerr, int) and yerr == 0):
        #    e = (y - LineOfBestFit.fitLine_straight(p, x)) / np.sqrt(yerr**2 + p[1]**2 * xerr**2)
        #else:
        #    e = y - predicted_y

        #return e
        return LineOfBestFit.__fitError_generic(LineOfBestFit.fitLine_straight, p, x, y, xerr, yerr, lambda p, x, y, xerr, yerr: np.sqrt(yerr**2 + p[1]**2 * xerr**2))

    @staticmethod
    def fitLine_quadratic(p: list, x):
        """
        Calculates a y value for a quadratic curve
        y_guess = ax^2 + bx + c

        p: [c, b, a]
        """
        f = p[2] * x**2 + p[1] * x + p[0]
        return f
    
    @staticmethod
    def fitError_quadratic(p: list, x, y, xerr, yerr):
        """
        Calculates the error for a quadratic curve
        error = (y - y_guess) / sqrt(dy^2 + (ax + b) dx^2)

        p: [c, b, a]
        """
        return LineOfBestFit.__fitError_generic(LineOfBestFit.fitLine_quadratic, p, x, y, xerr, yerr, lambda p, x, y, xerr, yerr: np.sqrt(yerr**2 + (p[2] * x + p[1])**2 * xerr**2))

    @staticmethod
    def fitLine_cubic(p: list, x):
        """
        Calculates a y value for a cubic curve
        y_guess = ax^3 + bx^2 + cx + d

        p: [d, c, b, a]
        """
        f = p[3] * x**3 + p[2] * x**2 + p[1] * x + p[0]
        return f
    
    @staticmethod
    def fitError_cubic(p: list, x, y, xerr, yerr):
        """
        Calculates the error for a cubic curve
        error = (y - y_guess) / sqrt(dy^2 + (ax^2 + bx + c) dx^2)

        p: [d, c, b, a]
        """
        return LineOfBestFit.__fitError_generic(LineOfBestFit.fitLine_cubic, p, x, y, xerr, yerr, lambda p, x, y, xerr, yerr: np.sqrt(yerr**2 + (p[3] * x**2 + p[2] * x + p[1])**2 * xerr**2))

    @staticmethod
    def fitLine_quartic(p: list, x):
        """
        Calculates a y value for a quartic curve
        y_guess = ax^4 + bx^3 + cx^2 + dx + e

        p: [e, d, c, b, a]
        """
        f = p[4] * x**4 + p[3] * x**3 + p[2] * x**2 + p[1] * x + p[0]
        return f
    
    @staticmethod
    def fitError_quartic(p: list, x, y, xerr, yerr):
        """
        Calculates the error for a quartic curve
        error = (y - y_guess) / sqrt(dy^2 + (ax^3 + bx^2 + cx + d) dx^2)

        p: [e, d, c, b, a]
        """
        return LineOfBestFit.__fitError_generic(LineOfBestFit.fitLine_quartic, p, x, y, xerr, yerr, lambda p, x, y, xerr, yerr: np.sqrt(yerr**2 + (p[4] * x**3 + p[3] * x**2 + p[2] * x + p[1])**2 * xerr**2))

    @staticmethod
    def fitLine_power(p: list, x):
        """
        Calculates a y value for a power curve
        y_guess = a x^b + c

        p: [c, b, a]
        """
        f = p[2] * x**p[1] + p[0]
        return f
    
    @staticmethod
    def fitError_power(p: list, x, y, xerr, yerr):
        """
        Calculates the error for a power curve
        error = (y - y_guess) / sqrt(dy^2 + a b x^(b-1) dx^2)

        p: [c, b, a]
        """
        return LineOfBestFit.__fitError_generic(LineOfBestFit.fitLine_power, p, x, y, xerr, yerr, lambda p, x, y, xerr, yerr: np.sqrt(yerr**2 + (p[2] * p[1] * x**(p[1] - 1)) * xerr**2))

    @staticmethod
    def fitLine_exponential(p: list, x):
        """
        Calculates a y value for an exponential curve (make b negitive for a logarithmic line)
        y_guess = a e^(bx) + c

        p: [c, b, a]
        """
        f = p[2] * np.exp(p[1] * x) + p[0]
        return f
    
    @staticmethod
    def fitError_exponential(p: list, x, y, xerr, yerr):
        """
        Calculates the error for an exponential curve (make b negitive for a logarithmic line)
        error = (y - y_guess) / sqrt(dy^2 + a b e^(bx) dx^2)

        p: [c, b, a]
        """
        return LineOfBestFit.__fitError_generic(LineOfBestFit.fitLine_exponential, p, x, y, xerr, yerr, lambda p, x, y, xerr, yerr: np.sqrt(yerr**2 + (p[2] * p[1] * np.exp(p[1] * x)) * xerr**2))

    @staticmethod
    def fitLine_sin(p: list, x):
        """
        Calculates a y value for a sine curve
        y_guess = a sin(bx + c)

        p: [c, b, a]
        """
        f = p[2] * np.sin(p[1] * x + p[0])
        return f
    
    @staticmethod
    def fitError_sin(p: list, x, y, xerr, yerr):
        """
        Calculates the error for a sine curve
        error = (y - y_guess) / sqrt(dy^2 + a b sin(bx + c) dx^2)

        p: [c, b, a]
        """
        #return LineOfBestFit.__fitError_generic(LineOfBestFit.fitLine_sin, p, x, y, xerr, yerr, lambda p, x, y, xerr, yerr: np.sqrt(yerr**2 + (p[2] * p[1] * np.cos(p[1] * x + p[0])) * xerr**2))
        LineOfBestFit.__fitError_generic(LineOfBestFit.fitLine_sin, p, x, y, xerr, yerr, lambda p, x, y, xerr, yerr: np.sqrt(yerr**2 + (p[2] * p[1] * np.cos(p[1] * x + p[0])) * xerr**2))

    @staticmethod
    def __fitError_generic(fitFunction, p: list, x, y, xerr, yerr, combineErrors = lambda p, x, y, xerr, yerr: np.sqrt(xerr**2 + yerr**2)):
        """
        Calculates the error for a specified function
        error = (y - y_guess) / e where by defualt e is: sqrt(dy^2 + dx^2)

        The error calculation should be e^2 = (δF/δx yerr)^2 + (δF/δy yerr)^2
                                            = (δF/δx yerr)^2 + yerr^2           as the diferential of F(x) with respect to y is 1
        """
        predicted_y = fitFunction(p, x)
        if not (isinstance(xerr, int) and xerr == 0) and not (isinstance(yerr, int) and yerr == 0):
            e = (y - predicted_y) / combineErrors(p, x, y, xerr, yerr)
        else:
            e = y - predicted_y
        return e

    @staticmethod
    def createGenericErrorFunc(fitFunction, combineErrors = lambda p, x, y, xerr, yerr: np.sqrt(xerr**2 + yerr**2)):
        """
        Wraps a fitting function inside the generic error function.
        Use this to utilise the generic fitting function with a custom fit.
        """
        return lambda p, x, y, xerr, yerr: __fitError_generic(fitFunction, p, x, y, xerr, yerr, combineErrors)



#def fitLine(p, x):
#    '''
#    Straight line
#    '''
#    f = p[0] + p[1]*x
#    return f
##
#def fitError(p, x, y, xerr, yerr):
#    '''
#    Error function for straight line fit
#    '''
#    e = (y - fitLine(p, x))/(np.sqrt(yerr**2 + p[1]**2*xerr**2))
#    return e
##
## Set initial values of fit parameters, run fit

#def plotStraightLineGraph(x, y, xErr, yErr, title, saveName = "fig.png", plotData = True, initialValues = [1.0, -1.0]):
#    """
#    returns: fitData, output, (mVal, mErr), (cVal, cErr), redchisq

#    Credits:
#        Tim Greenshaw
#    """
#    output = ""

#    nLen = np.size(x)

#    xData = x
#    yData = y
#    xError = xErr
#    yError = yErr

#    nPoints = nLen

#    pInit = initialValues
#    out = least_squares(fitError, pInit, args=(xData, yData, xError, yError))
#    #
#    fitOK = out.success
#    #
#    # Test if fit failed
#    if not fitOK:
#        output += "\n "
#        output += "\nFit failed"
#    else:
#        #
#        # get output
#        pFinal = out.x
#        cVal = pFinal[0]
#        mVal = pFinal[1]
#        #
#        #   Calculate chis**2 per point, summed chi**2 and chi**2/NDF
#        chiarr = fitError(pFinal, xData, yData, xError, yError)**2
#        chisq = np.sum(chiarr)
#        NDF = nPoints - 2
#        redchisq = chisq/NDF
#    #
#        np.set_printoptions(precision = 3)
#        output += "\n "
#        output += "\nFit quality:"
#        output += "\nchisq per point = \n " + str(chiarr)
#        output += "\nchisq = {:5.2f}, chisq/NDF = {:5.2f}.".format(chisq, redchisq)
#        output += "\n pFinal: "+str(pFinal)+" y: "+str(yData)
#        output += "\n dx: "+str(xError)+" dy: "+str(yError)
#        #
#        # Compute covariance
#        jMat = out.jac
#        jMat2 = np.dot(jMat.T, jMat)
#        detJmat2 = np.linalg.det(jMat2)
#        #
#        if detJmat2 < 1E-32:
#            output += "\nValue of determinat detJmat2 " + str(detJmat2)
#            output += "\nMatrix singular, error calculation failed."
#            output += "\n "
#            output += "\nParameters returned by fit:"
#            output += "\nIntercept = {:5.2f}".format(cVal)
#            output += "\nGradient = {:5.2f}".format(mVal)
#            output += "\n "
#            cErr = 0.0
#            mErr = 0.0
#        else:
#            covar = np.linalg.inv(jMat2)
#            cErr = np.sqrt(covar[0, 0])
#            mErr = np.sqrt(covar[1, 1])
#            #
#            output += "\n "
#            output += "\nParameters returned by fit:"
#            output += "\nIntercept (Sin theta i) = {:5.2f} +- {:5.2f}".format(cVal, cErr)
#            output += "\nGradient () = {:5.2f} +- {:5.2f}".format(mVal, mErr)
#            output += "\n "
#        #
#        # Calculate fitted function values
#        fitData = fitLine(pFinal, xData)
#        #
        
        
#    # Plot data
#    if plotData:
#        fig = plt.figure(figsize = (8, 6))
#        plt.title(title)
#        plt.xlabel('Radius squared (m^2)')
#        plt.ylabel('Period squared (s^2)')
#        plt.errorbar(xData, yData, xerr = xError, yerr = yError, fmt='r', \
#                    linestyle = '', label = "Data") 
#        plt.plot(xData, fitData, color = 'b', linestyle = '-', label = "Fit")
#        #plt.xlim(1.0, 7.0)
#        #plt.ylim(0.0, 3.0)
#        plt.grid(color = 'g')
#        plt.legend(loc = 2, prop={'size': 20})

#        plt.savefig(saveName)
#        plt.show()
    
#    return fitData, output, (mVal, mErr), (cVal, cErr), redchisq

from .figure import IAutomaticFigure


from ...edp import Event
from ._CardGroup import CardGroup
from ._PlayingCard import PlayingCard
from random import Random

class Deck(CardGroup):
    """
    A collection of 52+ unique playing cards. The exact number can vary dependant on the number of jokers added

    Atributes:
        str Name --> name of the collection
        int Jokers --> the number of jokers in this deck
        list[IPlayingCard] Cards --> the cards in the deck (including the jokers)
        QuasarCode.edp.Event ReturnCards --> request that all subscribers remove their held card objects

    __init__():
        str name --> name of the collection
        int jokers --> the number of jokers in this deck
    """
    def __init__(self, name: str, jokers: int = 0):
        self.Jokers: int = jokers
        self.Cards: list = []

        for suit in PlayingCard.AllowedSuits:
            if suit != PlayingCard.AllowedSuits.J:
                for value in PlayingCard.AllowedValues:
                    if value != PlayingCard.AllowedValues.Jo:
                        self.Cards.append(PlayingCard(value, suit))

        for i in range(jokers):
            self.Cards.append(PlayingCard(PlayingCard.AllowedValues.Jo, PlayingCard.AllowedSuits.J))

        super().__init__(name, self.Cards)

        self.__subscribers = []
        self.ReturnCards = Event()

    def InitialiseGroup(self, groups: list):
        """
        Initialises a card group for use with the deck

        Paramiters:
            list[CardGroup] groups --> the group(s) to be initialised
        """
        for group in groups:
            if group not in self.__subscribers:
                self.__subscribers.append(group)
                self.ReturnCards += group.ReturnCards

    def ReleaseGroup(self, groups: list):
        """
        Removes card groups from use with the deck

        Paramiters:
            list[CardGroup] groups --> the group(s) to remove
        """
        for group in groups:
            if group in self.__subscribers:
                self.__subscribers.remove(group)
                self.ReturnCards -= group.ReturnCards

    def Deal(self, hands: list, cardsPerHand: int = None, remainderGroups: list = None, suppressEmptyException: bool = False):
        """
        Either:
        1.) Deals the deck into a list of groups untill the deck is empty
            if only first paramiter is set

        2.) Deals a set number of cards from the deck into a list of groups and then deals the remaining cards to a list of groups if avalable

        Paramiters:
            list[CardGroup] hands --> The hands to deal into. Set only this paramiter for result (1)
            int cardsPerHand --> The number of cards to deal to each hand (deafult is None)
            list[CardGroup] remainderGroups --> The groups to deal any remaining cards into (deafult is None)
            bool suppressEmptyException --> Whether or not to suppress the exception generated when there isn't enough cards remaining to fill deal the number requested to each hand (deafult is False)
        """
        for hand in hands:
            if hand not in self.__subscribers:
                raise ValueError("One of the card groups provided wasn't initialised for use with this deck.")

        if remainderGroups is not None:
           for group in remainderGroups:
              if group not in self.__subscribers:
                raise ValueError("One of the remainder groups provided wasn't initialised for use with this deck.")

        randomiser = Random()

        if cardsPerHand == None:
            while len(self) > 0:
                for hand in hands:
                    if len(self) > 0:
                        hand.append(self.Remove(randomiser.randint(0, len(self))))
                    else:
                        break

        else:
            if not suppressEmptyException and cardsPerHand * lan(hands) > len(self):
                raise ValueError("There aren't enough cards in the deck to do this.")

            for i in range(cardsPerHand):
                for hand in hands:
                    if len(self) > 0:
                        hand.append(self.Remove(randomiser.randint(0, len(self))))
                    else:
                        break

                if len(self) == 0:
                    break

            if remainderGroups is not None:
                self.Deal(remainderGroups)

    def Reset(self):
        """
        Reset the deck so that it has all the cards it started with and asks all ICardGroups dealt to to remove references to the cards
        """
        self.clear()

        if len(self.__subscribers) > 0:
            self.ReturnCards.run(self)
        
        for card in self.Cards:
            self.append(card)


import math
from typing import Any, Literal, Optional

from matplotlib.figure import Figure
from matplotlib.axes import Axes
from matplotlib.axes._secondary_axes import SecondaryAxis

from ..Data._Rect import Rect
from ..IO.Caching._CacheableList import CacheableList
from ..Tools._Struct import CacheableStruct
from ..Tools._autoproperty import AutoProperty, AutoProperty_NonNullable
from ..Tools._CacheableFunction import CacheableFunction
from ._CachedPlotFontInfo import CachedPlotFontInfo
from ._CachedPlotCustomLegend import CachedPlotCustomLegend
from ._CachedPlotElements import CachedPlotElement, CachedPlotColourbar

class CachedPlot(CacheableStruct):
    title = AutoProperty[str](allow_uninitialised = True)
    plot_elements = AutoProperty_NonNullable[dict[str, CachedPlotElement]]()
    colourbars = AutoProperty_NonNullable[dict[str, CachedPlotColourbar]]()
    extent = AutoProperty_NonNullable[Rect]()
    aspect = AutoProperty[Literal["auto", "equal"]|float](allow_uninitialised = True)
    aspect_adjustable = AutoProperty[Literal["box", "datalim"]](allow_uninitialised = True)
    aspect_anchor = AutoProperty[Literal["C", "N", "NE", "E", "SE", "S", "SW", "W", "N"]|tuple[float, float]](allow_uninitialised = True)
    x_axis_label = AutoProperty[str](allow_uninitialised = True)
    y_axis_label = AutoProperty[str](allow_uninitialised = True)
    flip_x = AutoProperty_NonNullable[bool](default_value = False)
    flip_y = AutoProperty_NonNullable[bool](default_value = False)
    show_legend = AutoProperty_NonNullable[bool](default_value = False)
    legend_position = AutoProperty_NonNullable[Literal["best", "upper left", "upper center", "upper right", "center left", "center", "center right", "lower left", "lower center", "lower right"]|tuple[float, float]](default_value = "best")
    custom_legends = AutoProperty_NonNullable[dict[str, CachedPlotCustomLegend]]()
    show_x_ticks = AutoProperty_NonNullable[bool](default_value = True)
    show_y_ticks = AutoProperty_NonNullable[bool](default_value = True)
    show_x_ticks_on_other_side = AutoProperty_NonNullable[bool](default_value = False)
    show_y_ticks_on_other_side = AutoProperty_NonNullable[bool](default_value = False)
    show_x_tick_labels = AutoProperty_NonNullable[bool](default_value = True)
    show_y_tick_labels = AutoProperty_NonNullable[bool](default_value = True)
    show_x_tick_labels_on_other_side = AutoProperty_NonNullable[bool](default_value = False)
    show_y_tick_labels_on_other_side = AutoProperty_NonNullable[bool](default_value = False)
    x_ticks_inside = AutoProperty_NonNullable[bool](default_value = False)
    y_ticks_inside = AutoProperty_NonNullable[bool](default_value = False)
    alt_x_axis_functions = AutoProperty[CacheableList[CacheableFunction]](allow_uninitialised = True)
    alt_y_axis_functions = AutoProperty[CacheableList[CacheableFunction]](allow_uninitialised = True)
    alt_x_axis_label = AutoProperty[str](allow_uninitialised = True)
    alt_y_axis_label = AutoProperty[str](allow_uninitialised = True)
    show_alt_x_ticks = AutoProperty_NonNullable[bool](default_value = True)
    show_alt_y_ticks = AutoProperty_NonNullable[bool](default_value = True)
    show_alt_x_tick_labels = AutoProperty_NonNullable[bool](default_value = True)
    show_alt_y_tick_labels = AutoProperty_NonNullable[bool](default_value = True)
    alt_x_ticks_inside = AutoProperty_NonNullable[bool](default_value = False)
    alt_y_ticks_inside = AutoProperty_NonNullable[bool](default_value = False)
    default_font = AutoProperty_NonNullable[CachedPlotFontInfo]()
    title_font = AutoProperty_NonNullable[CachedPlotFontInfo]()
    x_axis_label_font = AutoProperty_NonNullable[CachedPlotFontInfo]()
    y_axis_label_font = AutoProperty_NonNullable[CachedPlotFontInfo]()
    alt_x_axis_label_font = AutoProperty_NonNullable[CachedPlotFontInfo]()
    alt_y_axis_label_font = AutoProperty_NonNullable[CachedPlotFontInfo]()
    x_tick_label_font = AutoProperty_NonNullable[CachedPlotFontInfo]()
    y_tick_label_font = AutoProperty_NonNullable[CachedPlotFontInfo]()
    alt_x_tick_label_font = AutoProperty_NonNullable[CachedPlotFontInfo]()
    alt_y_tick_label_font = AutoProperty_NonNullable[CachedPlotFontInfo]()

    _alt_x_axis = AutoProperty[SecondaryAxis](allow_uninitialised = True)
    _alt_y_axis = AutoProperty[SecondaryAxis](allow_uninitialised = True)

    def __init__(self, **kwargs):
        super().__init__(
            cacheable_attributes = (
                "title", "plot_elements", "colourbars", "extent", "aspect", "aspect_adjustable",
                "aspect_anchor", "x_axis_label", "y_axis_label", "flip_x", "flip_y", "show_legend",
                "legend_position", "custom_legends", "show_x_ticks", "show_y_ticks",
                "show_x_ticks_on_other_side", "show_y_ticks_on_other_side", "show_x_tick_labels",
                "show_y_tick_labels", "show_x_tick_labels_on_other_side",
                "show_y_tick_labels_on_other_side", "x_ticks_inside", "y_ticks_inside",
                "alt_x_axis_functions", "alt_y_axis_functions", "alt_x_axis_label",
                "alt_y_axis_label", "show_alt_x_ticks", "show_alt_y_ticks",
                "show_alt_x_tick_labels", "show_alt_y_tick_labels", "alt_x_ticks_inside",
                "alt_y_ticks_inside", "default_font", "title_font", "x_axis_label_font",
                "y_axis_label_font", "alt_x_axis_label_font", "alt_y_axis_label_font",
                "x_tick_label_font", "y_tick_label_font", "alt_x_tick_label_font",
                "alt_y_tick_label_font"
            ),
            **kwargs
        )
        self.plot_elements = {}
        self.colourbars = {}
        self.custom_legends = {}
        self.default_font = CachedPlotFontInfo()
        self.title_font = CachedPlotFontInfo()
        self.x_axis_label_font = CachedPlotFontInfo()
        self.y_axis_label_font = CachedPlotFontInfo()
        self.alt_x_axis_label_font = CachedPlotFontInfo()
        self.alt_y_axis_label_font = CachedPlotFontInfo()
        self.x_tick_label_font = CachedPlotFontInfo()
        self.y_tick_label_font = CachedPlotFontInfo()
        self.alt_x_tick_label_font = CachedPlotFontInfo()
        self.alt_y_tick_label_font = CachedPlotFontInfo()
    
    def __check_element_name_is_new(self, name: str) -> bool:
        return name not in self.plot_elements

    def add_element(self, name: str, element: CachedPlotElement) -> None:
        if not self.__check_element_name_is_new(name):
            raise KeyError(f"An element already exists with the name \"{name}\".")
        self.plot_elements[name] = element

    def add_colourbar(self, name: str, colourbar: CachedPlotColourbar) -> None:
        if name in self.colourbars:
            raise KeyError(f"A colourbar already exists with the name \"{name}\".")
        if self.__check_element_name_is_new(colourbar.target_element):
            raise KeyError(f"No element exists with the target name \"{colourbar.target_element}\".")
        self.colourbars[name] = colourbar

    def render(self, figure: Figure, axis: Axes, figure_default_font: Optional[CachedPlotFontInfo] = None, forward_kwargs: dict[str, dict[str, Any]]|None = None, forward_colourbar_kwargs: dict[str, dict[str, Any]]|None = None) -> None:
        if forward_kwargs is None:
            forward_kwargs = {}
        if forward_colourbar_kwargs is None:
            forward_colourbar_kwargs = {}
        for element in self.plot_elements:
            self.plot_elements[element].render(figure, axis, default_font = self.default_font.with_default(figure_default_font), **forward_kwargs.get(element, {}))

        for name, colourbar in self.colourbars.items():
            target: object
            if colourbar.target_element in self.plot_elements:
                target = self.plot_elements[colourbar.target_element]._result
            else:
                raise RuntimeError(f"Unable to locate target for colourbar \"{name}\".")
            colourbar.render(figure, axis, target, default_font = self.default_font.with_default(figure_default_font), **forward_colourbar_kwargs.get(name, {}))

        if self.title is not None:
            axis.set_title(self.title, **self.title_font.with_default(self.default_font).with_default(figure_default_font).fontdict)

        if self.x_axis_label is not None:
            axis.set_xlabel(self.x_axis_label, **self.x_axis_label_font.with_default(self.default_font).with_default(figure_default_font).fontdict)
        if self.y_axis_label is not None:
            axis.set_ylabel(self.y_axis_label, **self.y_axis_label_font.with_default(self.default_font).with_default(figure_default_font).fontdict)

        x_tick_label_font = self.x_tick_label_font.with_default(self.default_font).with_default(figure_default_font)
        x_tick_label_font_kwargs = {}
        if x_tick_label_font.size is not None:
            x_tick_label_font_kwargs["labelsize"] = x_tick_label_font.size
        if x_tick_label_font.colour is not None:
            x_tick_label_font_kwargs["labelcolor"] = x_tick_label_font.colour
        if x_tick_label_font.family is not None:
            x_tick_label_font_kwargs["labelfontfamily"] = x_tick_label_font.family
        axis.tick_params(
            axis            = "x",
            bottom          = self.show_x_ticks,
            top             = self.show_x_ticks_on_other_side and self.alt_x_axis_functions is None,
            direction       = "in" if self.x_ticks_inside else "out",
            labelbottom     = self.show_x_tick_labels,
            labeltop        = self.show_x_tick_labels_on_other_side,
            **x_tick_label_font_kwargs
        )

        y_tick_label_font = self.y_tick_label_font.with_default(self.default_font).with_default(figure_default_font)
        y_tick_label_font_kwargs = {}
        if y_tick_label_font.size is not None:
            y_tick_label_font_kwargs["labelsize"] = y_tick_label_font.size
        if y_tick_label_font.colour is not None:
            y_tick_label_font_kwargs["labelcolor"] = y_tick_label_font.colour
        if y_tick_label_font.family is not None:
            y_tick_label_font_kwargs["labelfontfamily"] = y_tick_label_font.family
        axis.tick_params(
            axis            = "y",
            left            = self.show_y_ticks,
            right           = self.show_y_ticks_on_other_side and self.alt_y_axis_functions is None,
            direction       = "in" if self.y_ticks_inside else "out",
            labelleft       = self.show_y_tick_labels,
            labelright      = self.show_y_tick_labels_on_other_side,
            **y_tick_label_font_kwargs
        )

        axis.set_xlim(self.extent.extent[0:2] if not self.flip_x else self.extent.extent[1::-1])
        axis.set_ylim(self.extent.extent[2:4] if not self.flip_y else self.extent.extent[3:1:-1])

        if self.alt_x_axis_functions is not None:
            self._alt_x_axis = axis.secondary_xaxis(
                "top",
                functions = tuple(self.alt_x_axis_functions)
            )
            if self.alt_x_axis_label is not None:
                self._alt_x_axis.set_xlabel(self.alt_x_axis_label, **self.alt_x_axis_label_font.with_default(self.default_font).with_default(figure_default_font).fontdict)
            alt_x_tick_label_font = self.alt_x_tick_label_font.with_default(self.x_tick_label_font).with_default(self.default_font).with_default(figure_default_font)
            alt_x_tick_label_font_kwargs = {}
            if alt_x_tick_label_font.size is not None:
                alt_x_tick_label_font_kwargs["labelsize"] = alt_x_tick_label_font.size
            if alt_x_tick_label_font.colour is not None:
                alt_x_tick_label_font_kwargs["labelcolor"] = alt_x_tick_label_font.colour
            if alt_x_tick_label_font.family is not None:
                alt_x_tick_label_font_kwargs["labelfontfamily"] = alt_x_tick_label_font.family
            self._alt_x_axis.tick_params(
                axis            = "x",
                top             = self.show_alt_x_ticks,
                direction       = "in" if self.alt_x_ticks_inside else "out",
                labeltop        = self.show_alt_x_tick_labels,
                **alt_x_tick_label_font_kwargs
            )

        if self.alt_y_axis_functions is not None:
            self._alt_y_axis = axis.secondary_yaxis(
                "right",
                functions = tuple(self.alt_y_axis_functions)
            )
            if self.alt_y_axis_label is not None:
                self._alt_y_axis.set_xlabel(self.alt_y_axis_label, **self.alt_y_axis_label_font.with_default(self.default_font).with_default(figure_default_font).fontdict)
            alt_y_tick_label_font = self.alt_y_tick_label_font.with_default(self.y_tick_label_font).with_default(self.default_font).with_default(figure_default_font)
            alt_y_tick_label_font_kwargs = {}
            if alt_y_tick_label_font.size is not None:
                alt_y_tick_label_font_kwargs["labelsize"] = alt_y_tick_label_font.size
            if alt_y_tick_label_font.colour is not None:
                alt_y_tick_label_font_kwargs["labelcolor"] = alt_y_tick_label_font.colour
            if alt_y_tick_label_font.family is not None:
                alt_y_tick_label_font_kwargs["labelfontfamily"] = alt_y_tick_label_font.family
            self._alt_y_axis.tick_params(
                axis            = "y",
                right           = self.show_alt_y_ticks,
                direction       = "in" if self.alt_y_ticks_inside else "out",
                labelright      = self.show_alt_y_tick_labels,
                **alt_y_tick_label_font_kwargs
            )

        if self.aspect is not None:
            axis.set_aspect(self.aspect, adjustable = self.aspect_adjustable, anchor = self.aspect_anchor)

        if self.show_legend:
            axis.legend(loc = self.legend_position)

        for legend in self.custom_legends.values():
            legend.render(figure, axis, default_font = self.default_font.with_default(figure_default_font), plot_elements = self.plot_elements)



from ._Bins import Bins
from ._CachedPlotFontInfo import CachedPlotFontInfo
from ._CachedPlotCustomLegend import CachedPlotCustomLegend, CachedPlotLegendBoxProperties
from ._CachedPlotElements import \
    CachedPlotElement,           \
    CachedPlotLine,              \
    CachedPlotScatter,           \
    CachedPlotErrorbar,          \
    CachedPlotHexbin,            \
    CachedPlotContour,           \
    CachedPlotColourbar,         \
    CachedPlotImage,             \
    CachedPlotText,              \
    CachedPlotPie
from ._CachedPlot import CachedPlot
from ._CachedFigureGrid import CachedFigureGrid
from ._CachedPlotFactory import CachedPlotFactory
from ._CachedFigureGridFactory import CachedFigureGridFactory
from ._Hexbin import Hexbin
from ._Contour import Contour



from abc import abstractmethod
from typing import Union, Collection, TypeVar, Type, Any

from ..IO.Caching._Cacheable import Cacheable

class Struct(object):
    """
    Structure type definition.
    Allows a number of public instance attributes to be declared and populated from the constructor.

    Constructor Design Order:
        1.) Declare all public attributes first.
        2.) super().__init__(**kwargs)

    To specify a human readable type name for errors, overload: _get_typename(self) -> Union[str, None]
    To specify type casts for attributes, set fields to an instance of TypeCastAutoProperty
    """
    
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            try:
                setattr(self, key, value)
            except AttributeError as e:
                raise AttributeError(f"{self._get_typename if self._get_typename is not None else type(self).__qualname__} struct defines no public field named \"{key}\".") from e

    @staticmethod
    def _get_typename() -> Union[str, None]:
        """
        Overload this to specify a string to appear in the error message provided when attempting to initalise a nonexistant attribute.
        """
        return None



T = TypeVar("T", bound = "CacheableStruct")

class CacheableStruct(Struct, Cacheable):
    """
    Structure type definition.
    Allows a number of public instance attributes to be declared and populated from the constructor.

    Constructor Design Order:
        1.) Declare all public attributes first.
        2.) super().__init__(cacheable_attributes = [<attribute-names>], **kwargs)
            where <attribute-names> is a list of the attributes that should be cached.

    To specify a human readable type name for errors, overload: _get_typename(self) -> Union[str, None]
    To specify type casts for attributes, set fields to an instance of TypeCastAutoProperty
    """
    
    def __init__(self, cacheable_attributes: Collection[str], **kwargs):
        self.__cacheable_attributes: tuple[str, ...] = tuple(cacheable_attributes)
        super().__init__(**kwargs)

    @classmethod
    def __from_cache_data__(cls: Type[T], data: dict[str, Any]) -> T:
        instance = cls()
        for attribute_name in data:
            if attribute_name == "__has_cacheables":
                continue
            attribute_data = data[attribute_name] if not isinstance(data[attribute_name], dict) or "datatype" not in data[attribute_name] else data[attribute_name]["datatype"].__from_cache_data__(data[attribute_name]["data"])
            if isinstance(attribute_data, dict) and "__has_cacheables" in attribute_data:
                loaded_attribute_data = attribute_data
                attribute_data = {}
                for sub_key in loaded_attribute_data:
                    if sub_key == "__has_cacheables":
                        continue#TODO: why is this needed again???
                    actual_subkey = sub_key if not isinstance(sub_key, dict) or "__key_needs_loading" not in loaded_attribute_data[sub_key] else sub_key["datatype"].__from_cache_data__(sub_key["data"])
                    actual_value: Any
                    if isinstance(loaded_attribute_data[sub_key], dict) and "datatype" in loaded_attribute_data[sub_key]:
                        actual_value = loaded_attribute_data[sub_key]["datatype"].__from_cache_data__(loaded_attribute_data[sub_key]["data"])
                    else:
                        actual_value = loaded_attribute_data[sub_key]
                    attribute_data[actual_subkey] = actual_value
                #attribute_data.pop("__has_cacheables")
            setattr(instance, attribute_name, attribute_data)
        return instance

    def __get_cache_data__(self) -> dict[str, Any]:
        data = {}
        for attribute_name in self.__cacheable_attributes:

            attribute_data = getattr(self, attribute_name)

            if isinstance(attribute_data, Cacheable):
                data[attribute_name] = { "datatype" : type(attribute_data), "data" : attribute_data.__get_cache_data__() }

            elif isinstance(attribute_data, dict):
                data_to_save: dict[Any, Any] = {}

                contains_cacheable = False

                for attr_data_key in attribute_data:
                    if isinstance(attr_data_key, Cacheable) or isinstance(attribute_data[attr_data_key], Cacheable):
                        contains_cacheable = True
                        key_is_cacheable = isinstance(attr_data_key, Cacheable)
                        key = attr_data_key if not key_is_cacheable else { "datatype": type(attr_data_key), "data": attr_data_key.__get_cache_data__() }
                        value = attribute_data[attr_data_key] if not isinstance(attribute_data[attr_data_key], Cacheable) else attribute_data[attr_data_key].__get_cache_data__()
                        data_to_save[key] = { "datatype": type(attribute_data[attr_data_key]), "data": value }
                        data_to_save[key]["__key_needs_loading"] = key_is_cacheable
                    else:
                        data_to_save[attr_data_key] = attribute_data[attr_data_key]

                if contains_cacheable:
                    data_to_save["__has_cacheables"] = True

                data[attribute_name] = data_to_save

            else:
                data[attribute_name] = attribute_data

        return data



from abc import abstractmethod
from decimal import Decimal
import math
from typing import Iterable, TypeVar, Generic

import numpy as np

from ._Rect import IRect, GenericRect, Rect, DiscreteRect

T = TypeVar("T")
class IRect3D(Generic[T]):
    """
    Interface for a 3D rectangle (cuboid) type.

    Use the following static classmethods on an implementation to create an instance:
        - create_from_size(x, y, z, width, height, depth)
        - create_from_centre(x, y, z, width, height, depth)
        - create_from_limits(x_min, x_max, y_min, y_max, z_min, z_max)
    """

    @classmethod
    @abstractmethod
    def create_from_size(cls, x: T, y: T, z: T, width: T, height: T, depth: T) -> "IRect3D[T]":
        """
        Creates an IRect3D instance using the specified position of the bottom-left-near corner and size.

        Parameters:
            T x:
                The x-coordinate of the bottom-left corner of the near plane.
            T y:
                The y-coordinate of the bottom-left corner of the near plane.
            T z:
                The z-coordinate of the near plane.
            T width:
                The width of the X-Y plane.
            T height:
                The height of the X-Y plane.
            T depth:
                The depth of the far plane from the near plane.

        Returns:
            IRect3D[T] -> A new IRect3D instance defined by the given position and size.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @classmethod
    @abstractmethod
    def create_from_centre(cls, x: T, y: T, z: T, width: T, height: T, depth: T) -> "IRect3D[T]":
        """
        Create a IRect3D instance using the centre coordinates and dimensions.

        Parameters:
            T x:
                The x-coordinate of the centre of the volume.
            T y:
                The y-coordinate of the centre of the volume.
            T z:
                The z-coordinate of the centre of the volume.
            T width:
                The width of the X-Y plane.
            T height:
                The height of the X-Y plane.
            T depth:
                The depth of the far plane from the near plane.

        Returns:
            IRect3D[T] -> Instance centred on (x, y, z) with the specified width, height, and depth.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @classmethod
    @abstractmethod
    def create_from_limits(cls, x_min: T, x_max: T, y_min: T, y_max: T, z_min: T, z_max: T) -> "IRect3D[T]":
        """
        Create a IRect3D instance from the specified boundary limits.
        This is the same as the constructor parameters and is implemented for consistency.

        Parameters:
            T x_min:
                The minimum x-coordinate of the rectangle.
            T x_max:
                The maximum x-coordinate of the rectangle.
            T y_min:
                The minimum y-coordinate of the rectangle.
            T y_max:
                The maximum y-coordinate of the rectangle.
            T z_min:
                The minimum z-coordinate of the rectangle.
            T z_max:
                The maximum z-coordinate of the rectangle.

        Returns:
            IRect3D[T] -> A new instance of the Rect3D class defined by the given limits.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @classmethod
    @abstractmethod
    def create_from_data(cls, x_data: Iterable[T], y_data: Iterable[T], z_data: Iterable[T]) -> "IRect3D[T]":
        """
        Create a IRect3D instance that encompasses the provided data.

        Parameters:
            Iterable[T] x_data:
                X-coordinates.
            Iterable[T] y_data:
                Y-coordinates.
            Iterable[T] z_data:
                Z-coordinates.

        Returns:
            IRect3D[T] -> A new instance of the Rect3D class that encompasses the data.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def x(self) -> T:
        """
        The x-coordinate of the bottom-left corner of the X-Y plane.

        Returns:
            T -> The x-coordinate of the X-Y plane's bottom-left corner.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @x.setter
    @abstractmethod
    def x(self, value: T) -> None:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def y(self) -> T:
        """
        The y-coordinate of the bottom-left corner of the X-Y plane.

        Returns:
            T -> The y-coordinate of the X-Y plane's bottom-left corner.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @y.setter
    @abstractmethod
    def y(self, value: T) -> None:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def z(self) -> T:
        """
        The z-coordinate of the near plane.

        Returns:
            T -> The z-coordinate of the near plane.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @z.setter
    @abstractmethod
    def z(self, value: T) -> None:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def x1(self) -> T:
        """
        The x-coordinate of the top-right corner of the X-Y plane.

        Returns:
            T -> The x-coordinate of the X-Y plane's top-right corner.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @x1.setter
    @abstractmethod
    def x1(self, value: T) -> None:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def y1(self) -> T:
        """
        The y-coordinate of the top-right corner of the X-Y plane.

        Returns:
            T -> The y-coordinate of the X-Y plane's top-right corner.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @y1.setter
    @abstractmethod
    def y1(self, value: T) -> None:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def z1(self) -> T:
        """
        The z-coordinate of the far plane.

        Returns:
            T -> The z-coordinate of the far plane.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @z1.setter
    @abstractmethod
    def z1(self, value: T) -> None:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def range(self) -> tuple[tuple[T, T], tuple[T, T], tuple[T, T]]:
        """
        The limits of the cuboid in the form ((x_min, x_max), (y_min, y_max), (z_min, z_max)).
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def width(self) -> T:
        """
        The width of the X-Y plane.

        Returns:
            T -> The width of the X-Y plane.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def height(self) -> T:
        """
        The height of the X-Y plane.

        Returns:
            T -> The height of the X-Y plane.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def depth(self) -> T:
        """
        The depth between the far to near planes.

        Returns:
            T -> The depth between the far to near planes.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def centre(self) -> tuple[T, T, T]:
        """
        The coordinates of the center of the cuboid.

        Returns:
            tuple[T, T, T] -> (x, y, z)
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def volume(self) -> T:
        """
        The volume of the cuboid.

        Returns:
            T -> The volume of the cuboid.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def plane_x_y(self) -> IRect[T]:
        """
        The X-Y plane of the cuboid (normal vector pointing in the Z direction).
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def plane_y_x(self) -> IRect[T]:
        """
        The Y-X plane of the cuboid (normal vector pointing in the -Z direction).
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def plane_x_z(self) -> IRect[T]:
        """
        The X-Z plane of the cuboid (normal vector pointing in the Y direction).
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def plane_z_x(self) -> IRect[T]:
        """
        The Z-X plane of the cuboid (normal vector pointing in the -Y direction).
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def plane_y_z(self) -> IRect[T]:
        """
        The Y-Z plane of the cuboid (normal vector pointing in the -X direction).
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def plane_z_y(self) -> IRect[T]:
        """
        The Z-Y plane of the cuboid (normal vector pointing in the X direction).
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def get_plane(self, axis_index: int = 2, flip: bool = False) -> IRect[T]:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __copy__(self) -> "IRect3D[T]":
        """
        Create a shallow copy of the cuboid.

        Returns:
            GenericIRect3D[T] -> A new instance of the GenericIRect3D class with the same properties.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __eq__(self, value) -> bool:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __ne__(self, value) -> bool:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __gt__(self, value) -> bool:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __ge__(self, value) -> bool:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __lt__(self, value) -> bool:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __le__(self, value) -> bool:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __add__(self, value: "IRect3D[T]") -> "IRect3D[T]":
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __iadd__(self, value: "IRect3D[T]") -> "IRect3D[T]":
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __sub__(self, value: "IRect3D[T]") -> "IRect3D[T]":
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __isub__(self, value: "IRect3D[T]") -> "IRect3D[T]":
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __mul__(self, value: float) -> "IRect3D[T]":
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __imul__(self, value: float) -> "IRect3D[T]":
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __div__(self, value: float) -> "IRect3D[T]":
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __idiv__(self, value: float) -> "IRect3D[T]":
        raise NotImplementedError("This method must be implemented by subclasses.")



class GenericRect3D(IRect3D[T]):
    """
    A 3D rectangle (cuboid) that supports coordinates using integers or floating point values.

    Parameters:
        T x_min:
            X-coordinate of the bottom-left corner of the X-Y plane.
        T x_max:
            X-coordinate of the top-right corner of the X-Y plane.
        T y_min:
            Y-coordinate of the bottom-left corner of the X-Y plane.
        T y_max:
            Y-coordinate of the top-right corner of the X-Y plane.
        T z_min:
            Z-coordinate of the near plane.
        T z_max:
            Z-coordinate of the far plane.

    Use the following classmethods to create an instance:
        - create_from_size(x, y, z, width, height, depth)
        - create_from_centre(x, y, z, width, height, depth)
        - create_from_limits(x_min, x_max, y_min, y_max, z_min, z_max)
    """

    def __init__(self, x_min: T, x_max: T, y_min: T, y_max: T, z_min: T, z_max: T) -> None:
        self.__x_min = x_min
        self.__x_max = x_max
        self.__y_min = y_min
        self.__y_max = y_max
        self.__z_min = z_min
        self.__z_max = z_max

    @classmethod
    def create_from_size(cls, x: T, y: T, z: T, width: T, height: T, depth: T) -> "GenericRect3D[T]":
        """
        Creates a GenericRect3D instance using the specified position of the bottom-left-near corner and size.

        Parameters:
            T x:
                The x-coordinate of the bottom-left corner of the near plane.
            T y:
                The y-coordinate of the bottom-left corner of the near plane.
            T z:
                The z-coordinate of the near plane.
            T width:
                The width of the X-Y plane.
            T height:
                The height of the X-Y plane.
            T depth:
                The depth of the far plane from the near plane.

        Returns:
            GenericRect3D[T] -> A new GenericRect3D instance defined by the given position and size.
        """
        return cls(x, x + width, y, y + height, z, z + depth)

    @classmethod
    def create_from_centre(cls, x: T, y: T, z: T, width: T, height: T, depth: T) -> "GenericRect3D[T]":
        """
        Create a GenericRect3D instance using the centre coordinates and dimensions.

        Parameters:
            T x:
                The x-coordinate of the centre of the volume.
            T y:
                The y-coordinate of the centre of the volume.
            T z:
                The z-coordinate of the centre of the volume.
            T width:
                The width of the X-Y plane.
            T height:
                The height of the X-Y plane.
            T depth:
                The depth of the far plane from the near plane.

        Returns:
            GenericRect3D[T] -> Instance centred on (x, y, z) with the specified width, height, and depth.
        """
        return cls(x - width / 2, x + width / 2, y - height / 2, y + height / 2, z - depth / 2, z + depth / 2)

    @classmethod
    def create_from_limits(cls, x_min: T, x_max: T, y_min: T, y_max: T, z_min: T, z_max: T) -> "GenericRect3D[T]":
        """
        Create a GenericRect3D instance from the specified boundary limits.
        This is the same as the constructor parameters and is implemented for consistency.

        Parameters:
            T x_min:
                The minimum x-coordinate of the rectangle.
            T x_max:
                The maximum x-coordinate of the rectangle.
            T y_min:
                The minimum y-coordinate of the rectangle.
            T y_max:
                The maximum y-coordinate of the rectangle.
            T z_min:
                The minimum z-coordinate of the rectangle.
            T z_max:
                The maximum z-coordinate of the rectangle.

        Returns:
            GenericRect3D[T] -> A new instance of the Rect3D class defined by the given limits.
        """
        return cls(x_min, x_max, y_min, y_max, z_min, z_max)

    @classmethod
    def create_from_data(cls, x_data: Iterable[T], y_data: Iterable[T], z_data: Iterable[T]) -> "GenericRect3D[T]":
        """
        Create a GenericRect3D instance that encompasses the provided data.

        Parameters:
            Iterable[T] x_data:
                X-coordinates.
            Iterable[T] y_data:
                Y-coordinates.
            Iterable[T] z_data:
                Z-coordinates.

        Returns:
            GenericRect3D[T] -> A new instance of the Rect3D class that encompasses the data.
        """
        return cls(min(x_data), max(x_data), min(y_data), max(y_data), min(z_data), max(z_data))

    @property
    def x(self) -> T:
        """
        The x-coordinate of the bottom-left corner of the X-Y plane.

        Returns:
            T -> The x-coordinate of the X-Y plane's bottom-left corner.
        """
        return self.__x_min

    @x.setter
    def x(self, value: T) -> None:
        self.__x_min = value

    @property
    def y(self) -> T:
        """
        The y-coordinate of the bottom-left corner of the X-Y plane.

        Returns:
            T -> The y-coordinate of the X-Y plane's bottom-left corner.
        """
        return self.__y_min

    @y.setter
    def y(self, value: T) -> None:
        self.__y_min = value

    @property
    def z(self) -> T:
        """
        The z-coordinate of the near plane.

        Returns:
            T -> The z-coordinate of the near plane.
        """
        return self.__z_min
    
    @z.setter
    def z(self, value: T) -> None:
        self.__z_min = value

    @property
    def x1(self) -> T:
        """
        The x-coordinate of the top-right corner of the X-Y plane.

        Returns:
            T -> The x-coordinate of the X-Y plane's top-right corner.
        """
        return self.__x_max

    @x1.setter
    def x1(self, value: T) -> None:
        self.__x_max = value

    @property
    def y1(self) -> T:
        """
        The y-coordinate of the top-right corner of the X-Y plane.

        Returns:
            T -> The y-coordinate of the X-Y plane's top-right corner.
        """
        return self.__y_max

    @y1.setter
    def y1(self, value: T) -> None:
        self.__y_max = value

    @property
    def z1(self) -> T:
        """
        The z-coordinate of the far plane.

        Returns:
            T -> The z-coordinate of the far plane.
        """
        return self.__z_max
    
    @z1.setter
    def z1(self, value: T) -> None:
        self.__z_max = value

    @property
    def range(self) -> tuple[tuple[T, T], tuple[T, T], tuple[T, T]]:
        """
        The limits of the cuboid in the form ((x_min, x_max), (y_min, y_max), (z_min, z_max)).
        """
        return (self.__x_min, self.__x_max), (self.__y_min, self.__y_max), (self.__z_min, self.__z_max)

    @property
    def width(self) -> T:
        """
        The width of the X-Y plane.

        Returns:
            T -> The width of the X-Y plane.
        """
        return self.__x_max - self.__x_min

    @property
    def height(self) -> T:
        """
        The height of the X-Y plane.
        
        Returns:
            T -> The height of the X-Y plane.
        """
        return self.__y_max - self.__y_min

    @property
    def depth(self) -> T:
        """
        The depth between the far to near planes.

        Returns:
            T -> The depth between the far to near planes.
        """
        return self.__z_max - self.__z_min

    @property
    def centre(self) -> tuple[T, T, T]:
        """
        The coordinates of the center of the cuboid.

        Returns:
            tuple[T, T, T] -> (x, y, z)
        """
        return ((self.__x_max + self.__x_min) / 2, (self.__y_max + self.__y_min) / 2, (self.__z_max + self.__z_min) / 2)

    @property
    def volume(self) -> T:
        """
        The volume of the cuboid.

        Returns:
            T -> The volume of the cuboid.
        """
        return self.width * self.height * self.depth

    @property
    def plane_x_y(self) -> GenericRect[T]:
        return GenericRect.create_from_limits(self.x, self.x1, self.y, self.y1)

    @property
    def plane_y_x(self) -> GenericRect[T]:
        return GenericRect.create_from_limits(self.y, self.y1, self.x, self.x1)

    @property
    def plane_x_z(self) -> GenericRect[T]:
        return GenericRect.create_from_limits(self.x, self.x1, self.z, self.z1)

    @property
    def plane_z_x(self) -> GenericRect[T]:
        return GenericRect.create_from_limits(self.z, self.z1, self.x, self.x1)

    @property
    def plane_y_z(self) -> GenericRect[T]:
        return GenericRect.create_from_limits(self.y, self.y1, self.z, self.z1)

    @property
    def plane_z_y(self) -> GenericRect[T]:
        return GenericRect.create_from_limits(self.z, self.z1, self.y, self.y1)

    def get_plane(self, axis_index: int = 2, flip: bool = False) -> GenericRect[T]:
        match axis_index:
            case 0:
                return self.plane_z_y if not flip else self.plane_y_z
            case 1:
                return self.plane_x_z if not flip else self.plane_z_x
            case 2:
                return self.plane_x_y if not flip else self.plane_y_x
            case _:
                raise ValueError("Axis must be 0 (X-Y), 1 (Y-Z), or 2 (X-Z).")

    def __copy__(self) -> "GenericRect3D[T]":
        """
        Create a shallow copy of the cuboid.

        Returns:
            GenericRect3D[T] -> A new instance of the GenericRect3D class with the same properties.
        """
        return GenericRect3D(self.__x_min, self.__x_max, self.__y_min, self.__y_max, self.__z_min, self.__z_max)

    def __str__(self) -> str:
        return f"Cuboid(x: {self.__x_min} -> {self.__x_max}, y: {self.__y_min} -> {self.__y_max}, z: {self.__z_min} -> {self.__z_max})"

    def __repr__(self) -> str:
        return "GenericRect3D: " + self.__str__()

    def __eq__(self, value):
        if isinstance(value, GenericRect3D):
            return self.__x_min == value.__x_min and self.__x_max == value.__x_max and self.__y_min == value.__y_min and self.__y_max == value.__y_max and self.__z_min == value.__z_min and self.__z_max == value.__z_max
        return False

    def __ne__(self, value) -> bool:
        return not self.__eq__(value)

    def __gt__(self, value) -> bool:
        if isinstance(value, IRect3D):
            return self.volume > value.volume
        elif isinstance(value, (int, float, Decimal, np.integer, np.floating)):
            return self.volume > value
        return False

    def __ge__(self, value) -> bool:
        if isinstance(value, IRect3D):
            return self.volume >= value.volume
        elif isinstance(value, (int, float, Decimal, np.integer, np.floating)):
            return self.volume >= value
        return False

    def __lt__(self, value) -> bool:
        if isinstance(value, IRect3D):
            return self.volume < value.volume
        elif isinstance(value, (int, float, Decimal, np.integer, np.floating)):
            return self.volume < value
        return False

    def __le__(self, value) -> bool:
        if isinstance(value, IRect3D):
            return self.volume <= value.volume
        elif isinstance(value, (int, float, Decimal, np.integer, np.floating)):
            return self.volume <= value
        return False

    def __add__(self, value: IRect3D[T]) -> "GenericRect3D[T]":
        return type(self).create_from_size(
            self.__x_min + value.x,
            self.__y_min + value.y,
            self.__z_min + value.z,
            self.width + value.width,
            self.height + value.height,
            self.depth + value.depth
        )

    def __iadd__(self, value: IRect3D[T]) -> "GenericRect3D[T]":
        current_width = self.width
        current_height = self.height
        current_depth = self.depth
        self.__x_min += value.x
        self.__y_min += value.y
        self.__z_min += value.z
        self.__x_max = self.__x_min + current_width + value.width
        self.__y_max = self.__y_min + current_height + value.height
        self.__z_max = self.__z_min + current_depth + value.depth
        return self

    def __sub__(self, value: IRect3D[T]) -> "GenericRect3D[T]":
        return type(self).create_from_size(
            self.__x_min - value.x,
            self.__y_min - value.y,
            self.__z_min - value.z,
            self.width - value.width,
            self.height - value.height,
            self.depth - value.depth
        )

    def __isub__(self, value: IRect3D[T]) -> "GenericRect3D[T]":
        current_width = self.width
        current_height = self.height
        current_depth = self.depth
        self.__x_min -= value.x
        self.__y_min -= value.y
        self.__z_min -= value.z
        self.__x_max = self.__x_min + current_width - value.width
        self.__y_max = self.__y_min + current_height - value.height
        self.__z_max = self.__z_min + current_depth - value.depth
        return self

    def __mul__(self, value: float) -> "GenericRect3D[T]":
        return type(self).create_from_size(
            self.__x_min,
            self.__y_min,
            self.__z_min,
            self.width * value,
            self.height * value,
            self.depth * value
        )

    def __imul__(self, value: float) -> "GenericRect3D[T]":
        self.__x_max = self.__x_min + (self.width * value)
        self.__y_max = self.__y_min + (self.height * value)
        self.__z_max = self.__z_min + (self.depth * value)
        return self

    def __div__(self, value: float) -> "GenericRect3D[T]":
        return type(self).create_from_size(
            self.__x_min,
            self.__y_min,
            self.__z_min,
            self.width / value,
            self.height / value,
            self.depth / value
        )

    def __idiv__(self, value: float) -> "GenericRect3D[T]":
        self.__x_max = self.__x_min + (self.width / value)
        self.__y_max = self.__y_min + (self.height / value)
        self.__z_max = self.__z_min + (self.depth / value)
        return self



class Rect3D(GenericRect3D[float]):
    """
    A 3D rectangle (cuboid).

    Parameters:
        float x_min:
            X-coordinate of the bottom-left corner of the X-Y plane.
        float x_max:
            X-coordinate of the top-right corner of the X-Y plane.
        float y_min:
            Y-coordinate of the bottom-left corner of the X-Y plane.
        float y_max:
            Y-coordinate of the top-right corner of the X-Y plane.
        float z_min:
            Z-coordinate of the near plane.
        float z_max:
            Z-coordinate of the far plane.

    Use the following classmethods to create an instance:
        - create_from_size(x, y, z, width, height, depth)
        - create_from_centre(x, y, z, width, height, depth)
        - create_from_limits(x_min, x_max, y_min, y_max, z_min, z_max)
    """

    @classmethod
    def create_from_size(cls, x: float, y: float, z: float, width: float, height: float, depth: float) -> "Rect3D":
        """
        Creates a Rect3D instance using the specified position of the bottom-left-near corner and size.

        Parameters:
            float x:
                The x-coordinate of the bottom-left corner of the near plane.
            float y:
                The y-coordinate of the bottom-left corner of the near plane.
            float z:
                The z-coordinate of the near plane.
            float width:
                The width of the X-Y plane.
            float height:
                The height of the X-Y plane.
            float depth:
                The depth of the far plane from the near plane.

        Returns:
            Rect3D -> A new Rect3D instance defined by the given position and size.
        """
        return super().create_from_size(x, width, y, height, z, depth)

    @classmethod
    def create_from_centre(cls, x: float, y: float, z: float, width: float, height: float, depth: float) -> "Rect3D":
        """
        Create a Rect3D instance using the centre coordinates and dimensions.

        Parameters:
            float x:
                The x-coordinate of the centre of the volume.
            float y:
                The y-coordinate of the centre of the volume.
            float z:
                The z-coordinate of the centre of the volume.
            float width:
                The width of the X-Y plane.
            float height:
                The height of the X-Y plane.
            float depth:
                The depth of the far plane from the near plane.

        Returns:
            Rect3D -> Instance centred on (x, y, z) with the specified width, height, and depth.
        """
        return super().create_from_centre(x, y, z, width, height, depth)

    @classmethod
    def create_from_limits(cls, x_min: float, x_max: float, y_min: float, y_max: float, z_min: float, z_max: float) -> "Rect3D":
        """
        Create a Rect3D instance from the specified boundary limits.
        This is the same as the constructor parameters and is implemented for consistency.

        Parameters:
            float x_min:
                The minimum x-coordinate of the rectangle.
            float x_max:
                The maximum x-coordinate of the rectangle.
            float y_min:
                The minimum y-coordinate of the rectangle.
            float y_max:
                The maximum y-coordinate of the rectangle.
            float z_min:
                The minimum z-coordinate of the rectangle.
            float z_max:
                The maximum z-coordinate of the rectangle.

        Returns:
            Rect3D -> A new instance of the Rect3D class defined by the given limits.
        """
        return super().create_from_limits(x_min, x_max, y_min, y_max, z_min, z_max)

    @classmethod
    def create_from_data(cls, x_data: Iterable[float], y_data: Iterable[float], z_data: Iterable[float]) -> "Rect3D":
        """
        Create a Rect3D instance that encompasses the provided data.

        Parameters:
            Iterable[T] x_data:
                X-coordinates.
            Iterable[T] y_data:
                Y-coordinates.
            Iterable[T] z_data:
                Z-coordinates.

        Returns:
            Rect3D -> A new instance of the Rect3D class that encompasses the data.
        """
        return super().create_from_data(x_data, y_data, z_data)

    @property
    def x(self) -> float:
        """
        The x-coordinate of the bottom-left corner of the X-Y plane.

        Returns:
            float -> The x-coordinate of the X-Y plane's bottom-left corner.
        """
        return GenericRect3D.x.fget(self)

    @x.setter
    def x(self, value: float) -> None:
        return GenericRect3D.x.fset(self, value)

    @property
    def y(self) -> float:
        """
        The y-coordinate of the bottom-left corner of the X-Y plane.

        Returns:
            float -> The y-coordinate of the X-Y plane's bottom-left corner.
        """
        return GenericRect3D.y.fget(self)

    @y.setter
    def y(self, value: float) -> None:
        return GenericRect3D.y.fset(self, value)

    @property
    def z(self) -> float:
        """
        The z-coordinate of the near plane.

        Returns:
            float -> The z-coordinate of the near plane.
        """
        return GenericRect3D.z.fget(self)

    @z.setter
    def z(self, value: float) -> None:
        return GenericRect3D.z.fset(self, value)

    @property
    def x1(self) -> float:
        """
        The x-coordinate of the top-right corner of the X-Y plane.

        Returns:
            float -> The x-coordinate of the X-Y plane's top-right corner.
        """
        return GenericRect3D.x1.fget(self)

    @x1.setter
    def x1(self, value: float) -> None:
        return GenericRect3D.x1.fset(self, value)

    @property
    def y1(self) -> float:
        """
        The y-coordinate of the top-right corner of the X-Y plane.

        Returns:
            float -> The y-coordinate of the X-Y plane's top-right corner.
        """
        return GenericRect3D.y1.fget(self)

    @y1.setter
    def y1(self, value: float) -> None:
        return GenericRect3D.y1.fset(self, value)

    @property
    def z1(self) -> float:
        """
        The z-coordinate of the top-right corner of the X-Y plane.

        Returns:
            float -> The z-coordinate of the X-Y plane's top-right corner.
        """
        return GenericRect3D.z1.fget(self)

    @z1.setter
    def z1(self, value: float) -> None:
        return GenericRect3D.z1.fset(self, value)

    @property
    def range(self) -> tuple[tuple[float, float], tuple[float, float], tuple[float, float]]:
        """
        The limits of the cuboid in the form ((x_min, x_max), (y_min, y_max), (z_min, z_max)).
        """
        return GenericRect3D.range.fget(self)

    @property
    def width(self) -> float:
        """
        The width of the X-Y plane.

        Returns:
            float -> The width of the X-Y plane.
        """
        return GenericRect3D.width.fget(self)

    @property
    def height(self) -> float:
        """
        The height of the X-Y plane.
        
        Returns:
            float -> The height of the X-Y plane.
        """
        return GenericRect3D.height.fget(self)

    @property
    def depth(self) -> float:
        """
        The depth between the far to near planes.

        Returns:
            float -> The depth between the far to near planes.
        """
        return GenericRect3D.depth.fget(self)

    @property
    def centre(self) -> tuple[float, float, float]:
        """
        The coordinates of the center of the cuboid.

        Returns:
            tuple[float, float, float] -> (x, y, z)
        """
        return GenericRect3D.centre.fget(self)

    @property
    def volume(self) -> float:
        """
        The volume of the cuboid.

        Returns:
            float -> The volume of the cuboid.
        """
        return  GenericRect3D.volume.fget(self)

    @property
    def plane_x_y(self) -> Rect:
        return GenericRect3D.plane_x_y.fget(self)

    @property
    def plane_y_x(self) -> Rect:
        return GenericRect3D.plane_y_x.fget(self)

    @property
    def plane_x_z(self) -> Rect:
        return GenericRect3D.plane_x_z.fget(self)

    @property
    def plane_z_x(self) -> Rect:
        return GenericRect3D.plane_z_x.fget(self)

    @property
    def plane_y_z(self) -> Rect:
        return GenericRect3D.plane_y_z.fget(self)

    @property
    def plane_z_y(self) -> Rect:
        return GenericRect3D.plane_z_y.fget(self)

    def get_plane(self, axis_index: int = 2, flip: bool = False) -> Rect:
        return GenericRect3D.get_plane(self, axis_index, flip)

    def __copy__(self) -> "Rect3D":
        """
        Create a shallow copy of the cuboid.

        Returns:
            Rect3D -> A new instance of the Rect3D class with the same properties.
        """
        ranges = self.range
        return Rect3D(ranges[0][0], ranges[0][1], ranges[1][0], ranges[1][1], ranges[2][0], ranges[2][1])

    def __repr__(self) -> str:
        return "Rect3D: " + self.__str__()

    def __add__(self, value: IRect3D[T]) -> "Rect3D":
        return super().__add__(value)

    def __sub__(self, value: IRect3D[T]) -> "Rect3D":
        return super().__sub__(value)

    def __mul__(self, value: float) -> "Rect3D":
        return super().__mul__(value)

    def __div__(self, value: float) -> "Rect3D":
        return super().__div__(value)



class DiscreteRect3D(GenericRect3D[int]):
    """
    A 3D rectangle (cuboid).

    Parameters:
        int x_min:
            X-coordinate of the bottom-left corner of the X-Y plane.
        int x_max:
            X-coordinate of the top-right corner of the X-Y plane.
        int y_min:
            Y-coordinate of the bottom-left corner of the X-Y plane.
        int y_max:
            Y-coordinate of the top-right corner of the X-Y plane.
        int z_min:
            Z-coordinate of the near plane.
        int z_max:
            Z-coordinate of the far plane.

    Use the following classmethods to create an instance:
        - create_from_size(x, y, z, width, height, depth)
        - create_from_centre(x, y, z, width, height, depth)
        - create_from_limits(x_min, x_max, y_min, y_max, z_min, z_max)
    """

    @classmethod
    def create_from_size(cls, x: int, y: int, z: int, width: int, height: int, depth: int) -> "DiscreteRect3D":
        """
        Creates a DiscreteRect3D instance using the specified position of the bottom-left-near corner and size.

        Parameters:
            int x:
                The x-coordinate of the bottom-left corner of the near plane.
            int y:
                The y-coordinate of the bottom-left corner of the near plane.
            int z:
                The z-coordinate of the near plane.
            int width:
                The width of the X-Y plane.
            int height:
                The height of the X-Y plane.
            int depth:
                The depth of the far plane from the near plane.

        Returns:
            DiscreteRect3D -> A new ReDiscreteRect3Dct3D instance defined by the given position and size.
        """
        return super().create_from_size(x, y, z, width, height, depth)
    
    @classmethod
    def create_from_centre(cls, x: int, y: int, z: int, width: int, height: int, depth: int) -> "DiscreteRect3D":
        """
        Create a DiscreteRect3D instance using the centre coordinates and dimensions.
        In the event that the width, height or depth are odd values, the upper bounds will be further from the centre by 1 then the lower bounds.

        Parameters:
            int x:
                The x-coordinate of the centre of the volume.
            int y:
                The y-coordinate of the centre of the volume.
            int z:
                The z-coordinate of the centre of the volume.
            int width:
                The width of the X-Y plane.
            int height:
                The height of the X-Y plane.
            int depth:
                The depth of the far plane from the near plane.

        Returns:
            DiscreteRect3D -> Instance centred on (x, y, z) with the specified width, height, and depth.
        """
        return cls(math.floor(x - width / 2), math.ceil(x + width / 2), math.floor(y - height / 2), math.ceil(y + height / 2), math.floor(z - depth / 2), math.ceil(z + depth / 2))

    @classmethod
    def create_from_limits(cls, x_min: int, x_max: int, y_min: int, y_max: int, z_min: int, z_max: int) -> "DiscreteRect3D":
        """
        Create a DiscreteRect3D instance from the specified boundary limits.
        This is the same as the constructor parameters and is implemented for consistency.

        Parameters:
            int x_min:
                The minimum x-coordinate of the rectangle.
            int x_max:
                The maximum x-coordinate of the rectangle.
            int y_min:
                The minimum y-coordinate of the rectangle.
            int y_max:
                The maximum y-coordinate of the rectangle.
            int z_min:
                The minimum z-coordinate of the rectangle.
            int z_max:
                The maximum z-coordinate of the rectangle.

        Returns:
            DiscreteRect3D -> A new instance of the DiscreteRect3D class defined by the given limits.
        """
        return super().create_from_limits(x_min, x_max, y_min, y_max, z_min, z_max)

    @classmethod
    def create_from_data(cls, x_data: Iterable[int], y_data: Iterable[int], z_data: Iterable[int]) -> "DiscreteRect3D":
        """
        Create a DiscreteRect3D instance that encompasses the provided data.

        Parameters:
            Iterable[int] x_data:
                X-coordinates.
            Iterable[int] y_data:
                Y-coordinates.
            Iterable[int] z_data:
                Z-coordinates.

        Returns:
            DiscreteRect3D -> A new instance of the DiscreteRect3D class that encompasses the data.
        """
        return super().create_from_data(x_data, y_data, z_data)

    @property
    def x(self) -> int:
        """
        The x-coordinate of the bottom-left corner of the X-Y plane.

        Returns:
            int -> The x-coordinate of the X-Y plane's bottom-left corner.
        """
        return GenericRect3D.x.fget(self)

    @x.setter
    def x(self, value: int) -> None:
        return GenericRect3D.x.fset(self, value)

    @property
    def y(self) -> int:
        """
        The y-coordinate of the bottom-left corner of the X-Y plane.

        Returns:
            int -> The y-coordinate of the X-Y plane's bottom-left corner.
        """
        return GenericRect3D.y.fget(self)

    @y.setter
    def y(self, value: int) -> None:
        return GenericRect3D.y.fset(self, value)
    
    @property
    def z(self) -> int:
        """
        The z-coordinate of the near plane.

        Returns:
            int -> The z-coordinate of the near plane.
        """
        return GenericRect3D.z.fget(self)

    @z.setter
    def z(self, value: int) -> None:
        return GenericRect3D.z.fset(self, value)

    @property
    def x1(self) -> int:
        """
        The x-coordinate of the top-right corner of the X-Y plane.

        Returns:
            int -> The x-coordinate of the X-Y plane's top-right corner.
        """
        return GenericRect3D.x1.fget(self)

    @x1.setter
    def x1(self, value: int) -> None:
        return GenericRect3D.x1.fset(self, value)

    @property
    def y1(self) -> int:
        """
        The y-coordinate of the top-right corner of the X-Y plane.

        Returns:
            int -> The y-coordinate of the X-Y plane's top-right corner.
        """
        return GenericRect3D.y1.fget(self)

    @y1.setter
    def y1(self, value: int) -> None:
        return GenericRect3D.y1.fset(self, value)

    @property
    def z1(self) -> int:
        """
        The z-coordinate of the top-right corner of the X-Y plane.

        Returns:
            int -> The z-coordinate of the X-Y plane's top-right corner.
        """
        return GenericRect3D.z1.fget(self)

    @z1.setter
    def z1(self, value: int) -> None:
        return GenericRect3D.z1.fset(self, value)

    @property
    def range(self) -> tuple[tuple[int, int], tuple[int, int], tuple[int, int]]:
        """
        The limits of the rectangle in the form ((x_min, x_max), (y_min, y_max), (z_min, z_max)).
        """
        return GenericRect3D.range.fget(self)

    @property
    def width(self) -> int:
        """
        The width of the X-Y plane.

        Returns:
            int -> The width of the X-Y plane.
        """
        return GenericRect3D.width.fget(self)

    @property
    def height(self) -> int:
        """
        The height of the X-Y plane.
        
        Returns:
            int -> The height of the X-Y plane.
        """
        return GenericRect3D.height.fget(self)

    @property
    def depth(self) -> int:
        """
        The depth between the far to near planes.

        Returns:
            int -> The depth between the far to near planes.
        """
        return GenericRect3D.depth.fget(self)

    @property
    def centre(self) -> tuple[int, int, int]:
        """
        The coordinates of the center of the cuboid.

        Returns:
            tuple[int, int, int] -> (x, y, z)
        """
        return (math.floor((self.x1 + self.x) / 2), math.floor((self.y1 + self.y) / 2), math.floor((self.z1 + self.z) / 2))

    @property
    def volume(self) -> int:
        """
        The volume of the cuboid.

        Returns:
            int -> The volume of the cuboid.
        """
        return  GenericRect3D.volume.fget(self)

    @property
    def plane_x_y(self) -> DiscreteRect:
        return GenericRect3D.plane_x_y.fget(self)

    @property
    def plane_y_x(self) -> DiscreteRect:
        return GenericRect3D.plane_y_x.fget(self)

    @property
    def plane_x_z(self) -> DiscreteRect:
        return GenericRect3D.plane_x_z.fget(self)

    @property
    def plane_z_x(self) -> DiscreteRect:
        return GenericRect3D.plane_z_x.fget(self)

    @property
    def plane_y_z(self) -> DiscreteRect:
        return GenericRect3D.plane_y_z.fget(self)

    @property
    def plane_z_y(self) -> DiscreteRect:
        return GenericRect3D.plane_z_y.fget(self)

    def get_plane(self, axis_index: int = 2, flip: bool = False) -> DiscreteRect:
        return GenericRect3D.get_plane(self, axis_index, flip)

    def __copy__(self) -> "DiscreteRect3D":
        """
        Create a shallow copy of the cuboid.

        Returns:
            DiscreteRect3D -> A new instance of the DiscreteRect3D class with the same properties.
        """
        ranges = self.range
        return DiscreteRect3D(ranges[0][0], ranges[0][1], ranges[1][0], ranges[1][1], ranges[2][0], ranges[2][1])

    def __repr__(self) -> str:
        return "DiscreteRect3D: " + self.__str__()

    def __add__(self, value: IRect3D[T]) -> "DiscreteRect3D":
        return super().__add__(value)

    def __sub__(self, value: IRect3D[T]) -> "DiscreteRect3D":
        return super().__sub__(value)

    def __mul__(self, value: float) -> "DiscreteRect3D":
        return type(self).create_from_size(
            self.x,
            self.y,
            self.z,
            math.floor(self.width * value),
            math.floor(self.height * value),
            math.floor(self.depth * value)
        )

    def __imul__(self, value: float) -> None:
        self.x1 = self.x + math.floor(self.width * value)
        self.y1 = self.y + math.floor(self.height * value)
        self.z1 = self.z + math.floor(self.depth * value)

    def __div__(self, value: float) -> "DiscreteRect3D":
        return type(self).create_from_size(
            self.x,
            self.y,
            self.z,
            math.floor(self.width / value),
            math.floor(self.height / value),
            math.floor(self.depth / value)
        )

    def __idiv__(self, value: float) -> None:
        self.x1 = self.x + math.floor(self.width / value)
        self.y1 = self.y + math.floor(self.height / value)
        self.z1 = self.z + math.floor(self.depth / value)



class UndersubscribedEventError(BaseException):
    """
    Indicates that there are two few subscribers to perform the requested operation.
    """
    def __init__(self, message):
        self.message = message


#from .fitline import LineType, LineOfBestFit
#from .graph import Graph
#from .figure import ManualFigure, Figure, DrawInstruction
from .data import consistancy_test, is_consistant, error_string, sig_fig, standard_form, standard_form_string, standard_form_string_if_nessessary, true_round
from ._plotting import polygon_y_error_region, polygon_x_error_region



from typing import Any, Optional, Literal

from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.artist import Artist
from matplotlib.typing import ColorType

from ..Data._Rect import Rect
from ..Tools._Struct import CacheableStruct
from ..Tools._autoproperty import AutoProperty, AutoProperty_NonNullable
from ._CachedPlotFontInfo import CachedPlotFontInfo
from ._CachedPlotElements import CachedPlotElement

class CachedPlotLegendBoxProperties(CacheableStruct):
    show             = AutoProperty[bool ](allow_uninitialised = True)
    round_corners    = AutoProperty[bool ](allow_uninitialised = True)
    shadow           = AutoProperty[bool ](allow_uninitialised = True)
    alpha            = AutoProperty[float](allow_uninitialised = True)
    face_colour      = AutoProperty[ColorType|Literal["inherit"]](allow_uninitialised = True)
    edge_colour      = AutoProperty[ColorType|Literal["inherit"]](allow_uninitialised = True)
    padding          = AutoProperty[float](allow_uninitialised = True) # in multiples of the font-size
    margin           = AutoProperty[float](allow_uninitialised = True) # in units of font-size
    def __init__(self, **kwargs):
        super().__init__(
            cacheable_attributes = ["show", "round_corners", "shadow", "alpha", "face_colour", "edge_colour", "padding", "margin"],
            **kwargs
        )

class CachedPlotCustomLegend(CacheableStruct):
    title                          = AutoProperty            [str                          ](allow_uninitialised = True)
    element_targets                = AutoProperty_NonNullable[dict[str,str|tuple[str,str]] ]() # Label : target element | (target plot, target element)
    reverse_order                  = AutoProperty_NonNullable[bool                         ](default_value = False)
    columns                        = AutoProperty_NonNullable[int                          ](default_value = 1)
    position                       = AutoProperty_NonNullable[Literal["best", "upper left", "upper center", "upper right", "center left", "center", "center right", "lower left", "lower center", "lower right"]|tuple[float, float]](default_value = "best")
    anchor_region                  = AutoProperty            [Rect                         ](allow_uninitialised = True) # Max. one of these \/ can be set
    anchor                         = AutoProperty            [tuple[int,int]               ](allow_uninitialised = True) # Max. one of these /\ can be set
    anchor_space                   = AutoProperty            [Literal["figure-space", "axis-space", "axis-data-space"]](allow_uninitialised = True)
    label_font                     = AutoProperty_NonNullable[CachedPlotFontInfo           ]()
    label_colours                  = AutoProperty            [str|Literal["linecolor", "markerfacecolor", "markeredgecolor"]|list[str|Literal["linecolor", "markerfacecolor", "markeredgecolor"]]](allow_uninitialised = True)
    title_font                     = AutoProperty_NonNullable[CachedPlotFontInfo           ]()
    alignment                      = AutoProperty_NonNullable[Literal["left", "center", "right"]](default_value = "center")

    line_render_points             = AutoProperty            [int                          ](allow_uninitialised = True)
    scatter_render_points          = AutoProperty            [int                          ](allow_uninitialised = True)
    scatter_render_points_y_offsets = AutoProperty_NonNullable[list[float]                  ](default_value = [0.375, 0.5, 0.3125]) # Relative to the font height (0 -> 1)
    marker_scale                   = AutoProperty            [float                        ](allow_uninitialised = True)
    marker_location                = AutoProperty_NonNullable[Literal["left", "right"]     ](default_value = "left")
    vertical_spacing               = AutoProperty            [float                        ](allow_uninitialised = True) # in units of font-size
    horizontal_spacing             = AutoProperty            [float                        ](allow_uninitialised = True) # in units of font-size
    handle_length                  = AutoProperty            [float                        ](allow_uninitialised = True) # in units of font-size
    handle_height                  = AutoProperty            [float                        ](allow_uninitialised = True) # in units of font-size
    handle_text_space              = AutoProperty            [float                        ](allow_uninitialised = True) # in units of font-size

    box                            = AutoProperty_NonNullable[CachedPlotLegendBoxProperties]()
    expand_into_horizontal_space   = AutoProperty_NonNullable[bool                         ](default_value = False)
    mouse_draggable                = AutoProperty_NonNullable[bool                         ](default_value = False)

    def __init__(self, **kwargs):
        super().__init__(
            cacheable_attributes = [
                "title", "element_targets", "reverse_order", "columns", "position",
                "anchor_region", "anchor", "anchor_space", "label_font", "label_colours",
                "title_font", "alignment", "line_render_points", "scatter_render_points",
                "scatter_render_points_y_offsets", "marker_scale", "marker_location",
                "vertical_spacing", "horizontal_spacing", "handle_length", "handle_height",
                "handle_text_space", "box", "expand_into_horizontal_space", "mouse_draggable"
            ],
            **kwargs
        )
        self.element_targets = {}
        self.label_font = CachedPlotFontInfo()
        self.title_font = CachedPlotFontInfo()
        self.box = CachedPlotLegendBoxProperties()

    def render(self, figure: Figure, axis: Axes|None, default_font: CachedPlotFontInfo, plot_elements: Optional[dict[str, CachedPlotElement]] = None, elements_by_figure_plot: Optional[dict[str, dict[str, CachedPlotElement]]] = None, *args: Any, **kwargs: Any) -> None:

        if plot_elements is None and elements_by_figure_plot is None:
            raise ValueError("At least one of plot_elements or elements_by_figure_plot must be provided to render a CachedPlotCustomLegend.")
        
        if self.anchor is None and self.anchor_region is not None:
            raise ValueError("Only one of anchor or anchor_region may be set.")
        
        if self.anchor_space is not None and self.anchor_space != "figure-space" and axis is None:
            raise ValueError("anchor_space cannot be set to a value other than 'figure-space' if axis is not provided.")
        
        anchor_transform = None if self.anchor_space is None else figure.transFigure if self.anchor_space == "figure-space" else axis.transAxes if self.anchor_space == "axis-space" else axis.transData

        artists = []
        labels = []
        for label, target in self.element_targets.items():
            target_plot: str|None
            target_element: str
            artist: Artist
            if isinstance(target, str):
                target_plot = None
                target_element = target
            else:
                target_plot = target[0]
                target_element = target[1]
            if target_plot is None:
                if plot_elements is None:
                    raise ValueError("plot_elements must be provided to render a CachedPlotCustomLegend with element_targets that do not specify a target plot.")
                artist = plot_elements[target_element]._result
            else:
                if elements_by_figure_plot is None:
                    raise ValueError("elements_by_figure_plot must be provided to render a CachedPlotCustomLegend with element_targets that specify a target plot.")
                artist = elements_by_figure_plot[target_plot][target_element]._result
            artists.append(artist)
            labels.append(label)

        (axis if axis is not None else figure).legend(
            handles              = artists,
            labels               = labels,
            loc                  = self.position,
            bbox_to_anchor       = self.anchor if self.anchor is not None else self.anchor_region.extent if self.anchor_region is not None else None,
            ncols                = self.columns,
            prop                 = self.label_font.with_default(default_font).fontproperties,
            labelcolor           = self.label_colours if self.label_colours is not None else self.label_font.with_default(default_font).colour if (self.label_font.colour is not None or default_font.colour is not None) else default_font.colour,
            numpoints            = self.line_render_points,
            scatterpoints        = self.scatter_render_points,
            scatteryoffsets      = self.scatter_render_points_y_offsets,
            markerscale          = self.marker_scale,
            markerfirst          = self.marker_location == "first",
            reverse              = self.reverse_order,
            frameon              = self.box.show,
            fancybox             = self.box.round_corners,
            shadow               = self.box.shadow,
            framealpha           = self.box.alpha,
            facecolor            = self.box.face_colour,
            edgecolor            = self.box.edge_colour,
            mode                 = "expand" if self.expand_into_horizontal_space else None,
            bbox_transform       = anchor_transform,
            title                = self.title,
            title_fontproperties = self.title_font.with_default(self.label_font).with_default(default_font).fontproperties,
            alignment            = self.alignment,
            borderpad            = self.box.padding,
            labelspacing         = self.vertical_spacing,
            handlelength         = self.handle_length,
            handleheight         = self.handle_height,
            handletextpad        = self.handle_text_space,
            borderaxespad        = self.box.margin,
            columnspacing        = self.horizontal_spacing,
#            handler_map          =,
            draggable            =self.mouse_draggable,
        )


from ..Data._Rect import Rect
from ._CachedFigureGrid import CachedFigureGrid
from ..IO.Caching._CacheTarget import CacheTarget

class CachedFigureGridFactory(object):
    def __init__(self, cache_directory: str) -> None:
        self.__cache_directory: str = cache_directory

    def new(self, extent: Rect) -> CachedFigureGrid:
        return CachedFigureGrid(extent = extent)

    def load(self, target: CacheTarget) -> CachedFigureGrid:
        return target.load_object(self.__cache_directory, CachedFigureGrid)

    def save(self, figure_grid: CachedFigureGrid, target: CacheTarget) -> None:
        target.save_object(self.__cache_directory, figure_grid)
